# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.support.ui import Select

from rbfWEB.base import LibraryComponent, keyword
from rbfWEB.utils import is_truthy, plural_or_not as s


class SelectElementKeywords(LibraryComponent):

    @keyword
    def get_list_items(self, locator, values=False):
        """Returns all labels or values of selection list ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        Returns visible labels by default, but values can be returned by
        setting the ``values`` argument to a true value (see `Boolean
        arguments`).

        Example:
        | ${labels} = | `Get List Items` | mylist              |             |
        | ${values} = | `Get List Items` | css:#example select | values=True |

        
        """
        options = self._get_options(locator)
        if is_truthy(values):
            return self._get_values(options)
        else:
            return self._get_labels(options)

    @keyword
    def get_selected_list_label(self, locator):
        """Returns label of selected option from selection list ``locator``.

        If there are multiple selected options, label of the first option
        is returned.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        select = self._get_select_list(locator)
        return select.first_selected_option.text

    @keyword
    def get_selected_list_labels(self, locator):
        """Returns labels of selected options from selection list ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        options = self._get_selected_options(locator)
        return self._get_labels(options)

    @keyword
    def get_selected_list_value(self, locator):
        """Returns value of selected option from selection list ``locator``.

        If there are multiple selected options, value of the first option
        is returned.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        select = self._get_select_list(locator)
        return select.first_selected_option.get_attribute('value')

    @keyword
    def get_selected_list_values(self, locator):
        """Returns values of selected options from selection list ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        options = self._get_selected_options(locator)
        return self._get_values(options)

    @keyword
    def list_selection_should_be(self, locator, *expected):
        """Verifies selection list ``locator`` has ``expected`` options selected.

        It is possible to give expected options both as visible labels and
        as values. 

        If no expected options are given, validates that the list has
        no selections. A more explicit alternative is using `List Should
        Have No Selections`.

        See the `Locating elements` section for details about the locator
        syntax.

        Examples:
        | `List Selection Should Be` | gender    | Female          |        |
        | `List Selection Should Be` | interests | Test Automation | Python |
        """
        self.info("Verifying list '%s' has option%s [ %s ] selected."
                  % (locator, s(expected), ' | '.join(expected)))
        self.page_should_contain_list(locator)
        options = self._get_selected_options(locator)
        labels = self._get_labels(options)
        values = self._get_values(options)
        if sorted(expected) not in [sorted(labels), sorted(values)]:
            raise AssertionError("List '%s' should have had selection [ %s ] "
                                 "but selection was [ %s ]."
                                 % (locator, ' | '.join(expected),
                                    self._format_selection(labels, values)))

    def _format_selection(self, labels, values):
        return ' | '.join('%s (%s)' % (label, value)
                          for label, value in zip(labels, values))

    @keyword
    def list_should_have_no_selections(self, locator):
        """Verifies selection list ``locator`` has no options selected.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Verifying list '%s' has no selections." % locator)
        options = self._get_selected_options(locator)
        if options:
            selection = self._format_selection(self._get_labels(options),
                                               self._get_values(options))
            raise AssertionError("List '%s' should have had no selection "
                                 "but selection was [ %s ]."
                                 % (locator, selection))

    @keyword
    def page_should_contain_list(self, locator, message=None, loglevel='TRACE'):
        """Verifies selection list ``locator`` is found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.assert_page_contains(locator, 'list', message, loglevel)

    @keyword
    def page_should_not_contain_list(self, locator, message=None, loglevel='TRACE'):
        """Verifies selection list ``locator`` is not found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.assert_page_not_contains(locator, 'list', message, loglevel)

    @keyword
    def select_all_from_list(self, locator):
        """Selects all options from multi-selection list ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Selecting all options from list '%s'." % locator)
        select = self._get_select_list(locator)
        if not select.is_multiple:
            raise RuntimeError("'Select All From List' works only with "
                               "multi-selection lists.")
        for i in range(len(select.options)):
            select.select_by_index(i)

    

    @keyword
    def select_from_list_by_index(self, locator, *indexes):
        """Selects options from selection list ``locator`` by ``indexes``.

        Indexes of list options start from 0.

        If more than one option is given for a single-selection list,
        the last value will be selected. With multi-selection lists all
        specified options are selected, but possible old selections are
        not cleared.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        if not indexes:
            raise ValueError("No indexes given.")
        self.info("Selecting options from selection list '%s' by index%s %s."
                  % (locator, '' if len(indexes) == 1 else 'es',
                     ', '.join(indexes)))
        select = self._get_select_list(locator)
        for index in indexes:
            select.select_by_index(int(index))

    @keyword
    def select_from_list_by_value(self, locator, *values):
        """Selects options from selection list ``locator`` by ``values``.

        If more than one option is given for a single-selection list,
        the last value will be selected. With multi-selection lists all
        specified options are selected, but possible old selections are
        not cleared.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        if not values:
            raise ValueError("No values given.")
        self.info("Selecting options from selection list '%s' by value%s %s."
                  % (locator, s(values), ', '.join(values)))
        select = self._get_select_list(locator)
        for value in values:
            select.select_by_value(value)

    @keyword
    def select_from_list_by_label(self, locator, *labels):
        """Selects options from selection list ``locator`` by ``labels``.

        If more than one option is given for a single-selection list,
        the last value will be selected. With multi-selection lists all
        specified options are selected, but possible old selections are
        not cleared.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        if not labels:
            raise ValueError("No labels given.")
        self.info("Selecting options from selection list '%s' by label%s %s."
                  % (locator, s(labels), ', '.join(labels)))
        select = self._get_select_list(locator)
        for label in labels:
            select.select_by_visible_text(label)

    @keyword
    def unselect_all_from_list(self, locator):
        """Unselects all options from multi-selection list ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        
        """
        self.info("Unselecting all options from list '%s'." % locator)
        select = self._get_select_list(locator)
        if not select.is_multiple:
            raise RuntimeError("Un-selecting options works only with "
                               "multi-selection lists.")
        select.deselect_all()

    

    @keyword
    def unselect_from_list_by_index(self, locator, *indexes):
        """Unselects options from selection list ``locator`` by ``indexes``.

        Indexes of list options start from 0. This keyword works only with
        multi-selection lists.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        if not indexes:
            raise ValueError("No indexes given.")
        self.info("Un-selecting options from selection list '%s' by index%s "
                  "%s." % (locator, '' if len(indexes) == 1 else 'es',
                           ', '.join(indexes)))
        select = self._get_select_list(locator)
        if not select.is_multiple:
            raise RuntimeError("Un-selecting options works only with "
                               "multi-selection lists.")
        for index in indexes:
            select.deselect_by_index(int(index))

    @keyword
    def unselect_from_list_by_value(self, locator, *values):
        """Unselects options from selection list ``locator`` by ``values``.

        This keyword works only with multi-selection lists.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        if not values:
            raise ValueError("No values given.")
        self.info("Un-selecting options from selection list '%s' by value%s "
                  "%s." % (locator, s(values), ', '.join(values)))
        select = self._get_select_list(locator)
        if not select.is_multiple:
            raise RuntimeError("Un-selecting options works only with "
                               "multi-selection lists.")
        for value in values:
            select.deselect_by_value(value)

    @keyword
    def unselect_from_list_by_label(self, locator, *labels):
        """Unselects options from selection list ``locator`` by ``labels``.

        This keyword works only with multi-selection lists.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        if not labels:
            raise ValueError("No labels given.")
        self.info("Un-selecting options from selection list '%s' by label%s "
                  "%s." % (locator, s(labels), ', '.join(labels)))
        select = self._get_select_list(locator)
        if not select.is_multiple:
            raise RuntimeError("Un-selecting options works only with "
                               "multi-selection lists.")
        for label in labels:
            select.deselect_by_visible_text(label)

    def _get_select_list(self, locator):
        el = self.find_element(locator, tag='list')
        return Select(el)

    def _get_options(self, locator):
        return self._get_select_list(locator).options

    def _get_selected_options(self, locator):
        return self._get_select_list(locator).all_selected_options

    def _get_labels(self, options):
        return [opt.text for opt in options]

    def _get_values(self, options):
        return [opt.get_attribute('value') for opt in options]



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .events import dispatch


class LibraryListener(object):
    ROBOT_LISTENER_API_VERSION = 2

    def start_suite(self, name, attrs):
        dispatch('scope_start', attrs['longname'])

    def end_suite(self, name, attrs):
        dispatch('scope_end', attrs['longname'])

    def start_test(self, name, attrs):
        dispatch('scope_start', attrs['longname'])

    def end_test(self, name, attrs):
        dispatch('scope_end', attrs['longname'])



import functools
import logging
import numbers
import os
import os.path as path
import re
import shutil
import string
import tempfile
from datetime import date, datetime, time
from enum import Enum
from random import choice

from xlrd import (XL_CELL_BLANK, XL_CELL_BOOLEAN, XL_CELL_DATE, XL_CELL_EMPTY,
                  XL_CELL_ERROR, XL_CELL_NUMBER, XL_CELL_TEXT)

LOGGER = logging.getLogger(__name__)


def random_temp_file(ext='txt'):
    return path.join(tempfile.gettempdir(), random_name() + '.' + ext)


def random_name():
    return ''.join(choice(string.ascii_letters) for x in range(8))


def is_file(file_path):
    return path.isfile(file_path)


def del_file(file_path):
    os.remove(file_path)


def copy_file(src, dest, force=False):
    if src == dest:
        raise ValueError('Same file error')
    if force and is_file(dest):
        del_file(dest)
    shutil.copy(src, dest)


def get_file_path(file_name):
    """
    Return None if file_name is None
    """
    if not file_name:
        return None
    if not file_name.endswith('.xlsx') and not file_name.endswith('.xls'):
        raise ValueError('Only support file with extenstion: xls and xlsx')
    file_path = path.normpath(file_name)
    if len(path.splitdrive(file_path)) > 1:
        return path.join(os.getcwd(), file_name)
    return file_name


def excel_name2coord(cell_name):
    matrix = list(filter(lambda x: x.strip(), re.split(r'(\d+)', cell_name.upper())))
    LOGGER.debug('Matrix: %s', matrix)
    if len(matrix) != 2 or not re.match(r'[A-Z]+', matrix[0]) or not re.match(r'\d+', matrix[1]):
        raise ValueError('Cell name is invalid')
    col = int(functools.reduce(lambda s, a: s * 26 + ord(a) - ord('A') + 1, matrix[0], 0)) - 1
    row = int(matrix[1]) - 1
    LOGGER.debug('Col, Row: %d, %d', col, row)
    return col, row



class DataType(Enum):

    DATE = XL_CELL_DATE
    TIME = XL_CELL_DATE * 11
    DATE_TIME = XL_CELL_DATE * 13
    TEXT = XL_CELL_TEXT
    NUMBER = XL_CELL_NUMBER
    CURRENCY = XL_CELL_NUMBER * 11
    PERCENTAGE = XL_CELL_NUMBER * 13
    BLANK = XL_CELL_BLANK
    EMPTY = XL_CELL_EMPTY
    ERROR = XL_CELL_ERROR
    BOOL = XL_CELL_BOOLEAN

    @staticmethod
    def is_date(dtype, value=None):
        if value and not dtype:
            return isinstance(value, (date, time))
        return DataType.DATE == dtype or DataType.TIME == dtype or DataType.DATE_TIME == dtype

    @staticmethod
    def is_number(ntype, value=None):
        if value and not ntype:
            return isinstance(value, numbers.Number)
        return DataType.NUMBER == ntype or DataType.CURRENCY == ntype or DataType.PERCENTAGE == ntype

    @staticmethod
    def is_bool(btype, value=None):
        if value and not btype:
            return isinstance(value, bool)
        return DataType.BOOL == btype

    @staticmethod
    def parse_type_by_value(type_value):
        return None if type_value is None else DataType(type_value)

    @staticmethod
    def parse_type(type_name):
        return None if type_name is None else DataType[type_name]


class DateFormat:

    @staticmethod
    def excel2python_format(dt_format):
        """
        https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior
        """
        if not dt_format:
            return ''
        _format = dt_format
        _format = re.sub(r'(?:(?<!%))y{4}', '%Y', _format)
        _format = re.sub(r'(?:(?<!%))y{2}', '%y', _format)
        _format = re.sub(r'(?:(?<!%))m{4}', '%B', _format)
        _format = re.sub(r'(?:(?<!%))m{3}', '%b', _format)
        _format = re.sub(r'(?:(?<!%))m{2}', '%m', _format)
        _format = re.sub(r'(?:(?<!%))m{1}', '%-m', _format)
        _format = re.sub(r'(?:(?<!%))d{2}', '%d', _format)
        _format = re.sub(r'(?:(?<!%))d{1}', '%-d', _format)
        hour_code = 'I' if re.search(r'\b(AM/PM)|(A/P)\b', _format, re.IGNORECASE) else 'H'
        _format = re.sub(r'\b(AM/PM)|(A/P)\b', '%p', _format)
        _format = re.sub(r'(?:(?<!%))H{2}', '%' + hour_code, _format)
        _format = re.sub(r'(?:(?<!%))H{1}', '%-' + hour_code, _format)
        _format = re.sub(r'(?:(?<!%))M{2}', '%M', _format)
        _format = re.sub(r'(?:(?<!%))M{1}', '%-M', _format)
        _format = re.sub(r'(?:(?<!%))S{2}', '%S', _format)
        _format = re.sub(r'(?:(?<!%))S{1}', '%-S', _format)
        return _format

    def __init__(self, date_format='yyyy-mm-dd', time_format='HH:MM:SS AM/PM', datetime_format='yyyy-mm-dd HH:MM'):
        self.date_format = date_format
        self.time_format = time_format
        self.datetime_format = datetime_format
        self.py_date_format = self.excel2python_format(self.date_format)
        self.py_time_format = self.excel2python_format(self.time_format)
        self.py_datetime_format = self.excel2python_format(self.datetime_format)
        self.datemode = 0  # 0: 1900-based, 1: 1904-based

    def get_excel_format(self, data_type, value):
        if value and isinstance(value, time) or DataType.TIME == data_type:
            return self.time_format.lower()
        if value and isinstance(value, datetime) or DataType.DATE_TIME == data_type:
            return self.datetime_format.lower()
        return self.date_format

    def get_py_format(self, data_type, value=None):
        if value and isinstance(value, time) or DataType.TIME == data_type:
            return self.py_time_format
        if value and isinstance(value, datetime) or DataType.DATE_TIME == data_type:
            return self.py_datetime_format
        return self.py_date_format

    def format(self, data_type, value):
        value = value.replace(year=max(1900 if self.datemode == 0 else 1904, value.year))
        return value.strftime(self.get_py_format(data_type))

    def parse(self, data_type, value):
        return value if isinstance(value, (datetime, time, date)) else datetime.strptime(str(value), self.get_py_format(data_type, value))


class NumberFormat:

    def __init__(self, decimal_sep='.', thousand_sep=',', precision='2'):
        self.number_format = '{0:' + thousand_sep + decimal_sep + precision + 'f}'

    def get_excel_format(self, data_type):
        # TODO: Missing percentage and currency
        return self.number_format

    def format(self, data_type, value):
        # TODO: Missing percentage and currency
        return self.number_format.format(value)

    def parse(self, data_type, value):
        return value if isinstance(value, numbers.Number) else float(value)


class BoolFormat:

    def __init__(self, bool_format='Yes/No'):
        self.true = bool_format.split('/')[0]
        self.false = bool_format.split('/')[1]

    def format(self, value):
        return self.true if value else self.false

    def parse(self, value):
        return value if isinstance(value, bool) else value and str(value).lower() == self.true.lower()



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from robot.utils import plural_or_not, secs_to_timestr, timestr_to_secs

from .deprecated import Deprecated
from .librarylistener import LibraryListener
from .seleniumversion import SELENIUM_VERSION
from .types import is_falsy, is_noney, is_string, is_truthy, PY3


def escape_xpath_value(value):
    if '"' in value and '\'' in value:
        parts_wo_apos = value.split('\'')
        return "concat('%s')" % "', \"'\", '".join(parts_wo_apos)
    if '\'' in value:
        return "\"%s\"" % value
    return "'%s'" % value



import logging
from operator import itemgetter

import natsort
from rbfExcel.utils import (BoolFormat, DataType, DateFormat, NumberFormat,
                              excel_name2coord, get_file_path, is_file)
from six import PY2
from xlrd import cellname, open_workbook, xldate

LOGGER = logging.getLogger(__name__)


class ExcelReader(object):

    def __init__(self, file_path, date_format=DateFormat(), number_format=NumberFormat(), bool_format=BoolFormat()):
        self.file_path = get_file_path(file_path)
        LOGGER.info('Opening file at %s', self.file_path)
        if not self.file_path or not is_file(self.file_path):
            self._workbook = None
            raise IOError('Excel file is not found') if PY2 else FileNotFoundError('Excel file is not found')
        self._workbook = open_workbook(self.file_path, formatting_info=self.is_xls, on_demand=True)
        self.date_format = date_format
        self.number_format = number_format
        self.bool_format = bool_format

    @property
    def is_xls(self):
        return not self.file_path.endswith('.xlsx')

    @property
    def extension(self):
        return 'xls' if self.is_xls else 'xlsx'

    @property
    def workbook(self):
        if not self._workbook:
            self._workbook = open_workbook(self.file_path, formatting_info=self.is_xls, on_demand=True)
        return self._workbook

    def _get_sheet(self, sheet_name):
        return self.workbook.sheet_by_name(sheet_name)

    def _get_cell_type(self, sheet_name, column, row):
        return self._get_sheet(sheet_name).cell_type(int(row), int(column))

    def get_sheet_names(self):
        """
        Returns the names of all the worksheets in the current workbook.
        """
        return self.workbook.sheet_names()

    def get_number_of_sheets(self):
        """
        Returns the number of worksheets in the current workbook.
        """
        return self.workbook.nsheets

    def get_column_count(self, sheet_name):
        """
        Returns the specific number of columns of the sheet name specified.
        """
        return self._get_sheet(sheet_name).ncols

    def get_row_count(self, sheet_name):
        """
        Returns the specific number of rows of the sheet name specified.
        """
        return self._get_sheet(sheet_name).nrows

    def get_column_values(self, sheet_name, column, include_empty_cells=True):
        """
        Returns the specific column values of the sheet name specified.
        """
        sheet = self._get_sheet(sheet_name)
        data = {}
        for row_index in range(sheet.nrows):
            cell = cellname(row_index, int(column))
            value = sheet.cell(row_index, int(column)).value
            data[cell] = value
        if not include_empty_cells:
            data = dict([(k, v) for (k, v) in data.items() if v])
        return natsort.natsorted(data.items(), key=itemgetter(0))

    def get_row_values(self, sheet_name, row, include_empty_cells=True):
        """
        Returns the specific row values of the sheet name specified.
        """
        sheet = self._get_sheet(sheet_name)
        data = {}
        for col_index in range(sheet.ncols):
            cell = cellname(int(row), col_index)
            value = sheet.cell(int(row), col_index).value
            data[cell] = value
        if not include_empty_cells:
            data = dict([(k, v) for (k, v) in data.items() if v])
        return natsort.natsorted(data.items(), key=itemgetter(0))

    def get_sheet_values(self, sheet_name, include_empty_cells=True):
        """
        Returns the values from the sheet name specified.
        """
        sheet = self._get_sheet(sheet_name)
        data = {}
        for row_index in range(sheet.nrows):
            for col_index in range(sheet.ncols):
                cell = cellname(row_index, col_index)
                value = sheet.cell(row_index, col_index).value
                data[cell] = value
        if not include_empty_cells:
            data = dict([(k, v) for (k, v) in data.items() if v])
        return natsort.natsorted(data.items(), key=itemgetter(0))

    def get_workbook_values(self, include_empty_cells=True):
        """
        Returns the values from each sheet of the current workbook.
        """
        sheet_data = []
        workbook_data = []
        for sheet_name in self.workbook.sheet_names():
            sheet_data = self.get_sheet_values(sheet_name, include_empty_cells)
            sheet_data.insert(0, sheet_name)
            workbook_data.append(sheet_data)
        return workbook_data

    def read_cell_data_by_name(self, sheet_name, cell_name, data_type=None, use_format=True):
        """
        Uses the cell name to return the data from that cell.
        """
        col, row = excel_name2coord(cell_name)
        return self.read_cell_data(sheet_name, col, row, data_type, use_format)
		
    def read_cell_data_by_row_and_column_header(self, sheet_name, row_header, column_header, data_type=None, use_format=True):
        """
        Uses the cell value to return the data from that cell.
        """
        sheet = self._get_sheet(sheet_name)
        row = sheet.row(0)
        for colidx, cell in enumerate(row):
            if cell.value == column_header :
                self.colidx = colidx
                        
        col = sheet.col(0)
        for rowidx, cell in enumerate(col):
            if cell.value == row_header :
                self.rowidx = rowidx				
        cell = sheet.cell(self.rowidx, self.colidx)
        return cell.value

    def read_cell_data(self, sheet_name, column, row, data_type=None, use_format=True):
        """
        Uses the column and row to return the data from that cell.

        :Args:
        data_type: Indicate explicit data type to convert
        use_format: Use format to convert data to string
        """
        sheet = self._get_sheet(sheet_name)
        cell = sheet.cell(int(row), int(column))
        ctype = DataType.parse_type_by_value(cell.ctype)
        value = cell.value
        gtype = DataType.parse_type(data_type)
        LOGGER.debug('Given Type: %s', gtype)
        LOGGER.debug('Cell Type: %s', ctype)
        LOGGER.debug('Cell Value: %s', value)
        if DataType.is_date(ctype):
            if gtype and not DataType.is_date(gtype):
                raise ValueError('Cell type does not match with given data type')
            date_value = xldate.xldate_as_datetime(value, self.date_format.datemode)
            if use_format:
                return self.date_format.format(gtype, date_value)
            elif DataType.DATE == gtype:
                return date_value.date()
            elif DataType.TIME == gtype:
                return date_value.time()
            return date_value
        if DataType.is_number(ctype):
            if gtype and not DataType.is_number(gtype):
                raise ValueError('Cell type does not match with given data type')
            return self.number_format.format(gtype, value) if use_format else value
        if DataType.is_bool(ctype):
            if gtype and not DataType.is_bool(gtype):
                raise ValueError('Cell type does not match with given data type')
            return self.bool_format.format(value) if use_format else value
        return value

    def check_cell_type(self, sheet_name, column, row, data_type):
        """
        Checks the type of value that is within the cell of the sheet name selected.
        """
        ctype = DataType.parse_type_by_value(self._get_cell_type(sheet_name, column, row))
        gtype = DataType.parse_type(data_type)
        LOGGER.debug('Given Type: %s', gtype)
        LOGGER.debug('Cell Type: %s', ctype)
        return ctype == gtype



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cv2
import numpy as np
from .utils import suppress_stderr
try:
    from gtk import gdk
except ImportError:
    gdk = None

try:
    from gi import require_version
    require_version('Gdk', '3.0')
    from gi.repository import Gdk
except ImportError:
    Gdk = None

try:
    from gi.repository import require_version
    require_version('Gdk', '3.0')
    from gi.repository import GdkPixbuf
except ImportError:
    GdkPixbuf = None


def _gtk_quality(format, quality):
    quality_setting = {}
    if format == 'png':
        quality_setting['compression'] = str(quality)
    else:
        quality_setting['quality'] = str(quality)
    return quality_setting


def _grab_screenshot_gtk_py2():
    window = gdk.get_default_root_window()
    if not window:
        raise RuntimeError('Taking screenshot failed.')
    width, height = window.get_size()
    pb = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, width, height)
    pb = pb.get_from_drawable(window, window.get_colormap(),
                              0, 0, 0, 0, width, height)
    if not pb:
        raise RuntimeError('Taking screenshot failed.')
    return pb


def _grab_screenshot_gtk_py3():
    window = Gdk.get_default_root_window()
    if not window:
        raise RuntimeError('Taking screenshot failed.')
    width = window.get_width()
    height = window.get_height()
    pb = Gdk.pixbuf_get_from_window(window, 0, 0, width, height)
    if not pb:
        raise RuntimeError('Taking screenshot failed.')
    return pb


def _grab_gtk_pb():
    if not gdk and not Gdk:
        raise RuntimeError('PyGTK not installed/supported on this platform.')
    if gdk:
        return _grab_screenshot_gtk_py2()
    elif Gdk:
        return _grab_screenshot_gtk_py3()


def _take_gtk_screenshot(path, format, quality):
    if not gdk and not Gdk:
        raise RuntimeError('PyGTK not installed/supported on this platform.')
    if gdk:
        return _take_gtk_screenshot_py2(path, format, quality)
    elif Gdk:
        return _take_gtk_screenshot_py3(path, format, quality)


def _take_gtk_screenshot_py2(path, format, quality):
    pb = _grab_screenshot_gtk_py2()
    quality_setting = _gtk_quality(format, quality)
    pb.save(path, format, quality_setting)
    return path


def _take_gtk_screenshot_py3(path, format, quality):
    pb = _grab_screenshot_gtk_py3()
    quality_setting = _gtk_quality(format, quality)
    pb.savev(path, format, [list(quality_setting.keys())[0]], [list(quality_setting.values())[0]])
    return path


def _take_gtk_screen_size():
    if not gdk and not Gdk:
        raise RuntimeError('PyGTK not installed/supported on this platform.')
    if gdk:
        window = gdk.get_default_root_window()
        if not window:
            raise RuntimeError('Taking screenshot failed.')
        width, height = window.get_size()
        return width, height
    elif Gdk:
        window = Gdk.get_default_root_window()
        if not window:
            raise RuntimeError('Taking screenshot failed.')
        width = window.get_width()
        height = window.get_height()
        return width, height


def _take_partial_gtk_screenshot(path, format, quality, left, top, width, height):
    if not gdk and not Gdk:
        raise RuntimeError('PyGTK not installed/supported on this platform.')
    if gdk:
        return _take_partial_gtk_screenshot_py2(path, format, quality, left, top, width, height)
    elif Gdk:
        return _take_partial_gtk_screenshot_py3(path, format, quality, left, top, width, height)


def _take_partial_gtk_screenshot_py2(path, format, quality, left, top, width, height):
    source_pb = _grab_screenshot_gtk_py2()
    quality_setting = _gtk_quality(format, quality)
    cropped_pb = source_pb.subpixbuf(left, top, width, height)
    if not cropped_pb:
        raise RuntimeError('Taking screenshot failed.')
    cropped_pb.save(path, format, quality_setting)
    return path


def _take_partial_gtk_screenshot_py3(path, format, quality, left, top, width, height):
    window = Gdk.get_default_root_window()
    if not window:
        raise RuntimeError('Taking screenshot failed.')
    cropped_pb = Gdk.pixbuf_get_from_window(window, left, top, width, height)
    if not cropped_pb:
        raise RuntimeError('Taking screenshot failed.')
    quality_setting = _gtk_quality(format, quality)
    cropped_pb.savev(path, format, [list(quality_setting.keys())[0]], [list(quality_setting.values())[0]])
    return path


def _record_gtk(path, fps, stop):
    if not gdk and not Gdk:
        raise RuntimeError('PyGTK not installed/supported on this platform.')
    if gdk:
        return _record_gtk_py2(path, fps, stop)
    elif Gdk:
        return _record_gtk_py3(path, fps, stop)


def _record_gtk_py2(path, fps, stop):
    window = gdk.get_default_root_window()
    fourcc = cv2.VideoWriter_fourcc(*'VP08')
    width, height = window.get_size()
    with suppress_stderr():
        vid = cv2.VideoWriter('%s' % path, fourcc, fps, (width, height))
    while not stop.isSet():
        pb = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, width, height)
        pb = pb.get_from_drawable(window, window.get_colormap(),
                                  0, 0, 0, 0, width, height)
        numpy_array = pb.get_pixels_array()
        frame = cv2.cvtColor(numpy_array, cv2.COLOR_RGB2BGR)
        vid.write(frame)
    vid.release()
    cv2.destroyAllWindows()


def _record_gtk_py3(path, fps, stop):
    window = Gdk.get_default_root_window()
    fourcc = cv2.VideoWriter_fourcc(*'VP08')
    width = window.get_width()
    height = window.get_height()
    with suppress_stderr():
        vid = cv2.VideoWriter('%s' % path, fourcc, fps, (width, height))
    while not stop.isSet():
        pb = Gdk.pixbuf_get_from_window(window, 0, 0, width, height)
        numpy_array = _convert_pixbuf_to_numpy(pb)
        frame = cv2.cvtColor(numpy_array,  cv2.COLOR_RGB2BGR)
        vid.write(frame)
    vid.release()
    cv2.destroyAllWindows()


def _convert_pixbuf_to_numpy(pixbuf):
    w, h, c, r = (pixbuf.get_width(), pixbuf.get_height(), pixbuf.get_n_channels(), pixbuf.get_rowstride())
    a = np.frombuffer(pixbuf.get_pixels(), dtype=np.uint8)
    if a.shape[0] == w * c * h:
        return a.reshape((h, w, c))
    else:
        b = np.zeros((h, w * c), 'uint8')
        for j in range(h):
            b[j, :] = a[r * j:r * j + w * c]
        return b.reshape((h, w, c))



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import namedtuple

from robot.utils import plural_or_not
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys


from rbfWEB.base import LibraryComponent, keyword
from rbfWEB.utils import (is_falsy, is_noney, is_truthy,
                                   plural_or_not as s)
from rbfWEB.errors import ElementNotFound


class ElementKeywords(LibraryComponent):

    @keyword(name='Get WebElement')
    def get_webelement(self, locator):
        """Returns the first WebElement matching the given ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        return self.find_element(locator)

    @keyword(name='Get WebElements')
    def get_webelements(self, locator):
        """Returns list of WebElement objects matching the ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        The keyword returns an empty list if there are no matching elements. 
        """
        return self.find_elements(locator)

    @keyword
    def element_should_contain(self, locator, expected, message=None, ignore_case=False):
        """Verifies that element ``locator`` contains text ``expected``.

        See the `Locating elements` section for details about the locator
        syntax.

        The ``message`` argument can be used to override the default error
        message.

        The ``ignore_case`` argument can be set to True to compare case
        insensitive, default is False.

        Use `Element Text Should Be` if you want to match the exact text,
        not a substring.
        """
        actual = actual_before = self.find_element(locator).text
        expected_before = expected
        if is_truthy(ignore_case):
            actual = actual.lower()
            expected = expected.lower()
        if expected not in actual:
            if is_noney(message):
                message = "Element '%s' should have contained text '%s' but "\
                          "its text was '%s'." % (locator, expected_before, actual_before)
            raise AssertionError(message)
        self.info("Element '%s' contains text '%s'." % (locator, expected_before))

    @keyword
    def element_should_not_contain(self, locator, expected, message=None, ignore_case=False):
        """Verifies that element ``locator`` does not contains text ``expected``.

        See the `Locating elements` section for details about the locator
        syntax.

        The ``message`` argument can be used to override the default error
        message.

        The ``ignore_case`` argument can be set to True to compare case
        insensitive, default is False.

        
        """
        actual = self.find_element(locator).text
        expected_before = expected
        if is_truthy(ignore_case):
            actual = actual.lower()
            expected = expected.lower()
        if expected in actual:
            if is_noney(message):
                message = "Element '%s' should not contain text '%s' but " \
                          "it did." % (locator, expected_before)
            raise AssertionError(message)
        self.info("Element '%s' does not contain text '%s'."
                  % (locator, expected_before))

    @keyword
    def page_should_contain(self, text, loglevel='TRACE'):
        """Verifies that current page contains ``text``.

        If this keyword fails, it automatically logs the page source
        using the log level specified with the optional ``loglevel``
        argument. Valid log levels are ``DEBUG``, ``INFO`` (default),
        ``WARN``, and ``NONE``. If the log level is ``NONE`` or below
        the current active log level the source will not be logged.
        """
        if not self._page_contains(text):
            self.ctx.log_source(loglevel)
            raise AssertionError("Page should have contained text '%s' "
                                 "but did not." % text)
        self.info("Current page contains text '%s'." % text)

    @keyword
    def page_should_contain_element(self, locator, message=None,
                                    loglevel='TRACE', limit=None):
        """Verifies that element ``locator`` is found on the current page.

        See the `Locating elements` section for details about the locator
        syntax.

        The ``message`` argument can be used to override the default error
        message.

        The ``limit`` argument can used to define how many elements the
        page should contain. When ``limit`` is ``None`` (default) page can
        contain one or more elements. When limit is a number, page must
        contain same number of elements.

        See `Page Should Contain` for explanation about the ``loglevel``
        argument.

        Examples assumes that locator matches to two elements.
        | `Page Should Contain Element` | div_name | limit=1    | # Keyword fails.                  |
        | `Page Should Contain Element` | div_name | limit=2    | # Keyword passes.                 |
        | `Page Should Contain Element` | div_name | limit=none | # None is considered one or more. |
        | `Page Should Contain Element` | div_name |            | # Same as above.                  |

        
        """
        if is_noney(limit):
            return self.assert_page_contains(locator, message=message,
                                             loglevel=loglevel)
        limit = int(limit)
        count = len(self.find_elements(locator))
        if count == limit:
            self.info('Current page contains {} element(s).'.format(count))
        else:
            if is_noney(message):
                message = ('Page should have contained "{}" element(s), '
                           'but it did contain "{}" element(s).'
                           .format(limit, count))
            self.ctx.log_source(loglevel)
            raise AssertionError(message)

    

    @keyword
    def page_should_not_contain(self, text, loglevel='TRACE'):
        """Verifies the current page does not contain ``text``.

        See `Page Should Contain` for explanation about the ``loglevel``
        argument.
        """
        if self._page_contains(text):
            self.ctx.log_source(loglevel)
            raise AssertionError("Page should not have contained text '%s'."
                                 % text)
        self.info("Current page does not contain text '%s'." % text)

    @keyword
    def page_should_not_contain_element(self, locator, message=None, loglevel='TRACE'):
        """Verifies that element ``locator`` is found on the current page.

        See the `Locating elements` section for details about the locator
        syntax.

        See `Page Should Contain` for explanation about ``message`` and
        ``loglevel`` arguments.
        """
        self.assert_page_not_contains(locator, message=message,
                                      loglevel=loglevel)

    @keyword
    def assign_id_to_element(self, locator, id):
        """Assigns temporary ``id`` to element specified by ``locator``.

        This is mainly useful if the locator is complicated and/or slow XPath
        expression and it is needed multiple times. Identifier expires when
        the page is reloaded.

        See the `Locating elements` section for details about the locator
        syntax.

        Example:
        | `Assign ID to Element` | //ul[@class='example' and ./li[contains(., 'Stuff')]] | my id |
        | `Page Should Contain Element` | my id |
        """
        self.info("Assigning temporary id '%s' to element '%s'." % (id, locator))
        element = self.find_element(locator)
        self.driver.execute_script("arguments[0].id = '%s';" % id, element)

    @keyword
    def element_should_be_disabled(self, locator):
        """Verifies that element identified with ``locator`` is disabled.

        This keyword considers also elements that are read-only to be
        disabled.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        if self.is_element_enabled(locator):
            raise AssertionError("Element '%s' is enabled." % locator)

    @keyword
    def element_should_be_enabled(self, locator):
        """Verifies that element identified with ``locator`` is enabled.

        This keyword considers also elements that are read-only to be
        disabled.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        if not self.is_element_enabled(locator):
            raise AssertionError("Element '%s' is disabled." % locator)

    @keyword
    def element_should_be_focused(self, locator):
        """Verifies that element identified with ``locator`` is focused.

        See the `Locating elements` section for details about the locator
        syntax.

        
        """
        element = self.find_element(locator)
        focused = self.driver.switch_to.active_element
        # Selenium 3.6.0 with Firefox return dict wich contains the selenium WebElement
        if isinstance(focused, dict):
            focused = focused['value']
        if element != focused:
            raise AssertionError("Element '%s' does not have focus." % locator)

    @keyword
    def element_should_be_visible(self, locator, message=None):
        """Verifies that the element identified by ``locator`` is visible.

        Herein, visible means that the element is logically visible, not
        optically visible in the current browser viewport. For example,
        an element that carries ``display:none`` is not logically visible,
        so using this keyword on that element would fail.

        See the `Locating elements` section for details about the locator
        syntax.

        The ``message`` argument can be used to override the default error
        message.
        """
        if not self.find_element(locator).is_displayed():
            if is_noney(message):
                message = ("The element '%s' should be visible, but it "
                           "is not." % locator)
            raise AssertionError(message)
        self.info("Element '%s' is displayed." % locator)

    @keyword
    def element_should_not_be_visible(self, locator, message=None):
        """Verifies that the element identified by ``locator`` is NOT visible.

        Passes if element does not exists. See `Element Should Be Visible`
        for more information about visibility and supported arguments.
        """
        element = self.find_element(locator, required=False)
        if element is None:
            self.info("Element '%s' did not exist." % locator)
        elif not element.is_displayed():
            self.info("Element '%s' exists but is not displayed." % locator)
        else:
            if is_noney(message):
                message = ("The element '%s' should not be visible, "
                           "but it is." % locator)
            raise AssertionError(message)

    @keyword
    def element_text_should_be(self, locator, expected, message=None, ignore_case=False):
        """Verifies that element ``locator`` contains exact text ``expected``.

        See the `Locating elements` section for details about the locator
        syntax.

        The ``message`` argument can be used to override the default error
        message.

        The ``ignore_case`` argument can be set to True to compare case
        insensitive, default is False.

        Use `Element Should Contain` if a substring match is desired.
        """
        self.info("Verifying element '%s' contains exact text '%s'."
                  % (locator, expected))
        text = before_text = self.find_element(locator).text
        if is_truthy(ignore_case):
            text = text.lower()
            expected = expected.lower()
        if text != expected:
            if is_noney(message):
                message = ("The text of element '%s' should have been '%s' "
                           "but it was '%s'."
                           % (locator, expected, before_text))
            raise AssertionError(message)

    @keyword
    def element_text_should_not_be(self, locator, not_expected, message=None, ignore_case=False):
        """Verifies that element ``locator`` does not contain exact text ``not_expected``.

        See the `Locating elements` section for details about the locator
        syntax.

        The ``message`` argument can be used to override the default error
        message.

        The ``ignore_case`` argument can be set to True to compare case
        insensitive, default is False.

        
        """
        self.info("Verifying element '%s' does not contains exact text '%s'."
                  % (locator, not_expected))
        text = self.find_element(locator).text
        before_not_expected = not_expected
        if is_truthy(ignore_case):
            text = text.lower()
            not_expected = not_expected.lower()
        if text == not_expected:
            if is_noney(message):
                message = ("The text of element '%s' was not supposed to be '%s'."
                           % (locator, before_not_expected))
            raise AssertionError(message)

    @keyword
    def get_element_attribute(self, locator, attribute):
        """Returns value of ``attribute`` from element ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        Example:
        | ${id}= | `Get Element Attribute` | css:h1 | id |

        
        """
        return self.find_element(locator).get_attribute(attribute)

    @keyword
    def element_attribute_value_should_be(self, locator, attribute, expected, message=None):
        """Verifies element identified by ``locator`` contains expected attribute value.

        See the `Locating elements` section for details about the locator
        syntax.

        Example:
        `Element Attribute Value Should Be` | css:img | href | value

        
        """
        current_expected = self.find_element(locator).get_attribute(attribute)
        if current_expected != expected:
            if is_noney(message):
                message = ("Element '%s' attribute should have value '%s' but "
                           "its value was '%s'." % (locator, expected, current_expected))
            raise AssertionError(message)
        self.info("Element '%s' attribute '%s' contains value '%s'." % (locator, attribute, expected))

    @keyword
    def get_horizontal_position(self, locator):
        """Returns horizontal position of element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        The position is returned in pixels off the left side of the page,
        as an integer.

        See also `Get Vertical Position`.
        """
        return self.find_element(locator).location['x']

    @keyword
    def get_element_size(self, locator):
        """Returns width and height of element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        Both width and height are returned as integers.

        Example:
        | ${width} | ${height} = | `Get Element Size` | css:div#container |
        """
        element = self.find_element(locator)
        return element.size['width'], element.size['height']

    @keyword
    def cover_element(self, locator):
        """Will cover elements identified by ``locator`` with a blue div without breaking page layout.
        
        See the `Locating elements` section for details about the locator
        syntax.
        
        
        
        Example:
        |`Cover Element` | css:div#container |
        """
        elements = self.find_elements(locator)
        if not elements:
            raise ElementNotFound("No element with locator '%s' found."
                                  % locator)
        for element in elements:
            script = """
old_element = arguments[0];
let newDiv = document.createElement('div');
newDiv.setAttribute("name", "covered");
newDiv.style.backgroundColor = 'blue';
newDiv.style.zIndex = '999';
newDiv.style.top = old_element.offsetTop + 'px';
newDiv.style.left = old_element.offsetLeft + 'px';
newDiv.style.height = old_element.offsetHeight + 'px';
newDiv.style.width = old_element.offsetWidth + 'px';
old_element.parentNode.insertBefore(newDiv, old_element);
old_element.remove();
newDiv.parentNode.style.overflow = 'hidden';
        """
            self.driver.execute_script(script, element)

    @keyword
    def get_value(self, locator):
        """Returns the value attribute of element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        return self.get_element_attribute(locator, 'value')

    @keyword
    def get_text(self, locator):
        """Returns the text value of element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        return self.find_element(locator).text

    @keyword
    def clear_element_text(self, locator):
        """Clears the value of text entry element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.find_element(locator).clear()

    @keyword
    def get_vertical_position(self, locator):
        """Returns vertical position of element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        The position is returned in pixels off the top of the page,
        as an integer.

        See also `Get Horizontal Position`.
        """
        return self.find_element(locator).location['y']

    @keyword
    def click_button(self, locator, modifier=False):
        """Clicks button identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, buttons are
        searched using ``id``, ``name`` and ``value``.

        See the `Click Element` keyword for details about the
        ``modifier`` argument.

        
        """
        if is_falsy(modifier):
            self.info("Clicking button '%s'." % locator)
            element = self.find_element(locator, tag='input', required=False)
            if not element:
                element = self.find_element(locator, tag='button')
            element.click()
        else:
            self._click_with_modifier(locator, ['button', 'input'], modifier)

    @keyword
    def click_image(self, locator, modifier=False):
        """Clicks an image identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, images are searched
        using ``id``, ``name``, ``src`` and ``alt``.

        See the `Click Element` keyword for details about the
        ``modifier`` argument.

        
        """
        if is_falsy(modifier):
            self.info("Clicking image '%s'." % locator)
            element = self.find_element(locator, tag='image', required=False)
            if not element:
                # A form may have an image as it's submit trigger.
                element = self.find_element(locator, tag='input')
            element.click()
        else:
            self._click_with_modifier(locator, ['image', 'input'], modifier)

    @keyword
    def click_link(self, locator, modifier=False):
        """Clicks a link identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, links are searched
        using ``id``, ``name``, ``href`` and the link text.

        See the `Click Element` keyword for details about the
        ``modifier`` argument.

        
        """
        if is_falsy(modifier):
            self.info("Clicking link '%s'." % locator)
            self.find_element(locator, tag='link').click()
        else:
            self._click_with_modifier(locator, ['link', 'link'], modifier)

    @keyword
    def click_element(self, locator, modifier=False):
        """Click element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        The ``modifier`` argument can be used to pass
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html#selenium.webdriver.common.keys.Keys|Selenium Keys]
        when clicking the element. The `+` can be used as a separator
        for different Selenium Keys. The `CTRL` is internally translated to
        `CONTROL` key. The ``modifier`` is space and case insensitive, example
        "alt" and " aLt " are supported formats to
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html#selenium.webdriver.common.keys.Keys.ALT|ALT key]
        . If ``modifier`` does not match to Selenium Keys, keyword fails.

        Example:
        | Click Element | id:button | | # Would click element without any modifiers. |
        | Click Element | id:button | CTRL | # Would click element with CTLR key pressed down. |
        | Click Element | id:button | CTRL+ALT | # Would click element with CTLR and ALT keys pressed down. |

        
        """
        if is_falsy(modifier):
            self.info("Clicking element '%s'." % locator)
            self.find_element(locator).click()
        else:
            self._click_with_modifier(locator, [None, None], modifier)

    def _click_with_modifier(self, locator, tag, modifier):
        self.info("Clicking %s '%s' with %s." % (tag if tag[0] else 'element', locator, modifier))
        modifier = self.parse_modifier(modifier)
        action = ActionChains(self.driver)
        for item in modifier:
            action.key_down(item)
        element = self.find_element(locator, tag=tag[0], required=False)
        if not element:
            element = self.find_element(locator, tag=tag[1])
        action.click(element)
        for item in modifier:
            action.key_up(item)
        action.perform()

    @keyword
    def click_element_at_coordinates(self, locator, xoffset, yoffset):
        """Click element ``locator`` at ``xoffset/yoffset``.

        Cursor is moved and the center of the element and x/y coordinates are
        calculated from that point.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Clicking element '%s' at coordinates x=%s, y=%s."
                  % (locator, xoffset, yoffset))
        element = self.find_element(locator)
        action = ActionChains(self.driver)
        action.move_to_element(element)
        action.move_by_offset(xoffset, yoffset)
        action.click()
        action.perform()

    @keyword
    def double_click_element(self, locator):
        """Double click element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Double clicking element '%s'." % locator)
        element = self.find_element(locator)
        action = ActionChains(self.driver)
        action.double_click(element).perform()

    @keyword
    def set_focus_to_element(self, locator):
        """Sets focus to element identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        
        """
        element = self.find_element(locator)
        self.driver.execute_script("arguments[0].focus();", element)

    

    @keyword
    def scroll_element_into_view(self, locator):
        """Scrolls an element identified by ``locator`` into view.

        See the `Locating elements` section for details about the locator
        syntax.

        
        """
        element = self.find_element(locator)
        ActionChains(self.driver).move_to_element(element).perform()

    @keyword
    def drag_and_drop(self, locator, target):
        """Drags element identified by ``locator`` into ``target`` element.

        The ``locator`` argument is the locator of the dragged element
        and the ``target`` is the locator of the target. See the
        `Locating elements` section for details about the locator syntax.

        Example:
        | `Drag And Drop` | css:div#element | css:div.target |
        """
        element = self.find_element(locator)
        target = self.find_element(target)
        action = ActionChains(self.driver)
        action.drag_and_drop(element, target).perform()

    @keyword
    def drag_and_drop_by_offset(self, locator, xoffset, yoffset):
        """Drags element identified with ``locator`` by ``xoffset/yoffset``.

        See the `Locating elements` section for details about the locator
        syntax.

        Element will be moved by ``xoffset`` and ``yoffset``, each of which
        is a negative or positive number specifying the offset.

        Example:
        | `Drag And Drop By Offset` | myElem | 50 | -35 | # Move myElem 50px right and 35px down |
        """
        element = self.find_element(locator)
        action = ActionChains(self.driver)
        action.drag_and_drop_by_offset(element, int(xoffset), int(yoffset))
        action.perform()

    @keyword
    def mouse_down(self, locator):
        """Simulates pressing the left mouse button on the element ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        The element is pressed without releasing the mouse button.

        See also the more specific keywords `Mouse Down On Image` and
        `Mouse Down On Link`.
        """
        self.info("Simulating Mouse Down on element '%s'." % locator)
        element = self.find_element(locator)
        action = ActionChains(self.driver)
        action.click_and_hold(element).perform()

    @keyword
    def mouse_out(self, locator):
        """Simulates moving mouse away from the element ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Simulating Mouse Out on element '%s'." % locator)
        element = self.find_element(locator)
        size = element.size
        offsetx = (size['width'] / 2) + 1
        offsety = (size['height'] / 2) + 1
        action = ActionChains(self.driver)
        action.move_to_element(element).move_by_offset(offsetx, offsety)
        action.perform()

    @keyword
    def mouse_over(self, locator):
        """Simulates hovering mouse over the element ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Simulating Mouse Over on element '%s'." % locator)
        element = self.find_element(locator)
        action = ActionChains(self.driver)
        action.move_to_element(element).perform()

    @keyword
    def mouse_up(self, locator):
        """Simulates releasing the left mouse button on the element ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Simulating Mouse Up on element '%s'." % locator)
        element = self.find_element(locator)
        ActionChains(self.driver).release(element).perform()

    @keyword
    def open_context_menu(self, locator):
        """Opens context menu on element identified by ``locator``."""
        element = self.find_element(locator)
        action = ActionChains(self.driver)
        action.context_click(element).perform()

    @keyword
    def simulate_event(self, locator, event):
        """Simulates ``event`` on element identified by ``locator``.

        This keyword is useful if element has ``OnEvent`` handler that
        needs to be explicitly invoked.

        See the `Locating elements` section for details about the locator
        syntax.

        
        """
        element = self.find_element(locator)
        script = """
element = arguments[0];
eventName = arguments[1];
if (document.createEventObject) { // IE
    return element.fireEvent('on' + eventName, document.createEventObject());
}
var evt = document.createEvent("HTMLEvents");
evt.initEvent(eventName, true, true);
return !element.dispatchEvent(evt);
        """
        self.driver.execute_script(script, element, event)

    
    @keyword
    def press_keys(self, locator=None, *keys):
        """Simulates user pressing key(s) to an element or on the active browser.


        If ``locator`` evaluates as false, see `Boolean arguments` for more
        details, then the ``keys`` are sent to the currently active browser.
        Otherwise element is searched and ``keys`` are send to the element
        identified by the ``locator``. In later case, keyword fails if element
        is not found. See the `Locating elements` section for details about
        the locator syntax.

        ``keys`` arguments can contain one or many strings, but it can not
        be empty. ``keys`` can also be a combination of
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html|Selenium Keys]
        and strings or a single Selenium Key. If Selenium Key is combined
        with strings, Selenium key and strings must be separated by the
        `+` character, like in `CONTROL+c`. Selenium Keys
        are space and case sensitive and Selenium Keys are not parsed
        inside of the string. Example AALTO, would send string `AALTO`
        and `ALT` not parsed inside of the string. But `A+ALT+O` would
        found Selenium ALT key from the ``keys`` argument. It also possible
        to press many Selenium Keys down at the same time, example
        'ALT+ARROW_DOWN`.

        If Selenium Keys are detected in the ``keys`` argument, keyword
        will press the Selenium Key down, send the strings and
         then release the Selenium Key. If keyword needs to send a Selenium
        Key as a string, then each character must be separated with
        `+` character, example `E+N+D`.

        `CTRL` is alias for
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html#selenium.webdriver.common.keys.Keys.CONTROL|Selenium CONTROL]
        and ESC is alias for
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html#selenium.webdriver.common.keys.Keys.ESCAPE|Selenium ESCAPE]

        

        Examples:
        | `Press Keys` | text_field | AAAAA          |            | # Sends string "AAAAA" to element.                                                |
        | `Press Keys` | None       | BBBBB          |            | # Sends string "BBBBB" to currently active browser.                               |
        | `Press Keys` | text_field | E+N+D          |            | # Sends string "END" to element.                                                  |
        | `Press Keys` | text_field | XXX            | YY         | # Sends strings "XXX" and "YY" to element.                                        |
        | `Press Keys` | text_field | XXX+YY         |            | # Same as above.                                                                  |
        | `Press Keys` | text_field | ALT+ARROW_DOWN |            | # Pressing "ALT" key down, then pressing ARROW_DOWN and then releasing both keys. |
        | `Press Keys` | text_field | ALT            | ARROW_DOWN | # Pressing "ALT" key and then pressing ARROW_DOWN.                                |
        | `Press Keys` | text_field | CTRL+c         |            | # Pressing CTRL key down, sends string "c" and then releases CTRL key.            |
        | `Press Keys` | button     | RETURN         |            | # Pressing "ENTER" key to element.                                                |
        """
        parsed_keys = self._parse_keys(*keys)
        if is_truthy(locator):
            self.info('Sending key(s) %s to %s element.' % (keys, locator))
        else:
            self.info('Sending key(s) %s to page.' % str(keys))
        self._press_keys(locator, parsed_keys)

    def _press_keys(self, locator, parsed_keys):
        if is_truthy(locator):
            element = self.find_element(locator)
        else:
            element = None
        for parsed_key in parsed_keys:
            actions = ActionChains(self.driver)
            special_keys = []
            for key in parsed_key:
                if self._selenium_keys_has_attr(key.original):
                    special_keys = self._press_keys_special_keys(actions, element, parsed_key,
                                                                 key, special_keys)
                else:
                    self._press_keys_normal_keys(actions, element, key)
            for special_key in special_keys:
                self.info('Releasing special key %s.' % special_key.original)
                actions.key_up(special_key.converted)
            actions.perform()

    def _press_keys_normal_keys(self, actions, element, key):
        self.info('Sending key%s %s' % (plural_or_not(key.converted), key.converted))
        if element:
            actions.send_keys_to_element(element, key.converted)
        else:
            actions.send_keys(key.converted)

    def _press_keys_special_keys(self, actions, element, parsed_key, key, special_keys):
        if len(parsed_key) == 1 and element:
            self.info('Pressing special key %s to element.' % key.original)
            actions.send_keys_to_element(element, key.converted)
        elif len(parsed_key) == 1 and not element:
            self.info('Pressing special key %s to browser.' % key.original)
            actions.send_keys(key.converted)
        else:
            self.info('Pressing special key %s down.' % key.original)
            actions.key_down(key.converted)
            special_keys.append(key)
        return special_keys

    @keyword
    def get_all_links(self):
        """Returns a list containing ids of all links found in current page.

        If a link has no id, an empty string will be in the list instead.
        """
        links = self.find_elements("tag=a")
        return [link.get_attribute('id') for link in links]

    @keyword
    def mouse_down_on_link(self, locator):
        """Simulates a mouse down event on a link identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, links are searched
        using ``id``, ``name``, ``href`` and the link text.
        """
        element = self.find_element(locator, tag='link')
        action = ActionChains(self.driver)
        action.click_and_hold(element).perform()

    @keyword
    def page_should_contain_link(self, locator, message=None, loglevel='TRACE'):
        """Verifies link identified by ``locator`` is found from current page.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, links are searched
        using ``id``, ``name``, ``href`` and the link text.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.
        """
        self.assert_page_contains(locator, 'link', message, loglevel)

    @keyword
    def page_should_not_contain_link(self, locator, message=None, loglevel='TRACE'):
        """Verifies link identified by ``locator`` is not found from current page.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, links are searched
        using ``id``, ``name``, ``href`` and the link text.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.
        """
        self.assert_page_not_contains(locator, 'link', message, loglevel)

    @keyword
    def mouse_down_on_image(self, locator):
        """Simulates a mouse down event on an image identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, images are searched
        using ``id``, ``name``, ``src`` and ``alt``.
        """
        element = self.find_element(locator, tag='image')
        action = ActionChains(self.driver)
        action.click_and_hold(element).perform()

    @keyword
    def page_should_contain_image(self, locator, message=None, loglevel='TRACE'):
        """Verifies image identified by ``locator`` is found from current page.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, images are searched
        using ``id``, ``name``, ``src`` and ``alt``.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.
        """
        self.assert_page_contains(locator, 'image', message, loglevel)

    @keyword
    def page_should_not_contain_image(self, locator, message=None, loglevel='TRACE'):
        """Verifies image identified by ``locator`` is found from current page.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, images are searched
        using ``id``, ``name``, ``src`` and ``alt``.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.
        """
        self.assert_page_not_contains(locator, 'image', message, loglevel)

    

    @keyword
    def get_element_count(self, locator):
        """Returns number of elements matching ``locator``.

        If you wish to assert the number of matching elements, use
        `Page Should Contain Element` with ``limit`` argument. Keyword will
        always return an integer.

        Example:
        | ${count} =       | `Get Element Count` | name:div_name  |
        | `Should Be True` | ${count} > 2        |                |

        
        """
        return len(self.find_elements(locator))

    @keyword
    def add_location_strategy(self, strategy_name, strategy_keyword, persist=False):
        """Adds a custom location strategy.

        See `Custom locators` for information how to create and use
        custom strategies. `Remove Location Strategy` can be used to
        remove a registered strategy.

        Location strategies are automatically removed after leaving the
        current scope by default. Setting ``persist`` to a true value (see
        `Boolean arguments`) will cause the location strategy to stay
        registered throughout the life of the test.
        """
        self.element_finder.register(strategy_name, strategy_keyword, persist)

    @keyword
    def remove_location_strategy(self, strategy_name):
        """Removes a previously added custom location strategy.

        See `Custom locators` for information how to create and use
        custom strategies.
        """
        self.element_finder.unregister(strategy_name)

    def _map_ascii_key_code_to_key(self, key_code):
        map = {
            0: Keys.NULL,
            8: Keys.BACK_SPACE,
            9: Keys.TAB,
            10: Keys.RETURN,
            13: Keys.ENTER,
            24: Keys.CANCEL,
            27: Keys.ESCAPE,
            32: Keys.SPACE,
            42: Keys.MULTIPLY,
            43: Keys.ADD,
            44: Keys.SEPARATOR,
            45: Keys.SUBTRACT,
            56: Keys.DECIMAL,
            57: Keys.DIVIDE,
            59: Keys.SEMICOLON,
            61: Keys.EQUALS,
            127: Keys.DELETE
        }
        key = map.get(key_code)
        if key is None:
            key = chr(key_code)
        return key

    def _map_named_key_code_to_special_key(self, key_name):
        try:
            return getattr(Keys, key_name)
        except AttributeError:
            message = "Unknown key named '%s'." % (key_name)
            self.debug(message)
            raise ValueError(message)

    def _page_contains(self, text):
        self.driver.switch_to.default_content()

        if self.is_text_present(text):
            return True

        subframes = self.find_elements("xpath://frame|//iframe")
        self.debug('Current frame has %d subframes.' % len(subframes))
        for frame in subframes:
            self.driver.switch_to.frame(frame)
            found_text = self.is_text_present(text)
            self.driver.switch_to.default_content()
            if found_text:
                return True
        return False

    def parse_modifier(self, modifier):
        modifier = modifier.upper()
        modifiers = modifier.split('+')
        keys = []
        for item in modifiers:
            item = item.strip()
            item = self._parse_aliases(item)
            if hasattr(Keys, item):
                keys.append(getattr(Keys, item))
            else:
                raise ValueError("'%s' modifier does not match to Selenium Keys"
                                 % item)
        return keys

    def _parse_keys(self, *keys):
        if not keys:
            raise AssertionError('"keys" argument can not be empty.')
        list_keys = []
        for key in keys:
            separate_keys = self._separate_key(key)
            separate_keys = self._convert_special_keys(separate_keys)
            list_keys.append(separate_keys)
        return list_keys

    def _parse_aliases(self, key):
        if key == 'CTRL':
            return 'CONTROL'
        if key == 'ESC':
            return 'ESCAPE'
        return key

    def _separate_key(self, key):
        one_key = ''
        list_keys = []
        for char in key:
            if char == '+' and one_key != '':
                list_keys.append(one_key)
                one_key = ''
            else:
                one_key += char
        if one_key:
            list_keys.append(one_key)
        return list_keys

    def _convert_special_keys(self, keys):
        KeysRecord = namedtuple('KeysRecord', 'converted, original')
        converted_keys = []
        for key in keys:
            key = self._parse_aliases(key)
            if self._selenium_keys_has_attr(key):
                converted_keys.append(KeysRecord(getattr(Keys, key), key))
            else:
                converted_keys.append(KeysRecord(key, key))
        return converted_keys

    def _selenium_keys_has_attr(self, key):
        try:
            return hasattr(Keys, key)
        except UnicodeError:  # To support Python 2 and non ascii characters.
            return False



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .common import RobotStatements

class Keyword(RobotStatements):
    '''A robotframework keyword

    A keyword is identical to a testcase in almost all respects
    except for some of the metadata it supports (which this definition
    doesn't (yet) account for...).
    '''
    def __init__(self, parent, linenumber, name):
        RobotStatements.__init__(self)
        self.linenumber = linenumber
        self.name = name
        self.rows = []
        self.parent = parent

    def __repr__(self):
        # should this return the fully qualified name?
        return "<Keyword: %s>" % self.name




# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from robot.utils import get_link_path

from rbfWEB.base import LibraryComponent, keyword
from rbfWEB.utils import is_noney


class ScreenshotKeywords(LibraryComponent):

    @keyword
    def set_screenshot_directory(self, path):
        """Sets the directory for captured screenshots.

        ``path`` argument specifies the absolute path to a directory where
        the screenshots should be written to. If the directory does not
        exist, it will be created. The directory can also be set when
        `importing` the library. If it is not configured anywhere,
        screenshots are saved to the same directory where Robot Framework's
        log file is written.

        ``persist`` argument is deprecated and has no effect.

        The previous value is returned and can be used to restore
        the original value later if needed.

        
        """
        if is_noney(path):
            path = None
        else:
            path = os.path.abspath(path)
            self._create_directory(path)
        previous = self.ctx.screenshot_root_directory
        self.ctx.screenshot_root_directory = path
        return previous

    @keyword
    def capture_page_screenshot(self, filename='selenium-screenshot-{index}.png'):
        """Takes screenshot of the current page and embeds it into log file.

        ``filename`` argument specifies the name of the file to write the
        screenshot into. The directory where screenshots are saved can be
        set when `importing` the library or by using the `Set Screenshot
        Directory` keyword. If the directory is not configured, screenshots
        are saved to the same directory where Robot Framework's log file is
        written.

        If ``filename`` contains marker
        ``{index}``, it will be automatically replaced with unique running
        index preventing files to be overwritten. Indices start from 1,
        and how they are represented can be customized using Python's
        [https://docs.python.org/2/library/string.html#formatstrings|
        format string syntax].

        An absolute path to the created screenshot file is returned.

        Examples:
        | `Capture Page Screenshot` |                                        |
        | `File Should Exist`       | ${OUTPUTDIR}/selenium-screenshot-1.png |
        | ${path} =                 | `Capture Page Screenshot`              |
        | `File Should Exist`       | ${OUTPUTDIR}/selenium-screenshot-2.png |
        | `File Should Exist`       | ${path}                                |
        | `Capture Page Screenshot` | custom_name.png                        |
        | `File Should Exist`       | ${OUTPUTDIR}/custom_name.png           |
        | `Capture Page Screenshot` | custom_with_index_{index}.png          |
        | `File Should Exist`       | ${OUTPUTDIR}/custom_with_index_1.png   |
        | `Capture Page Screenshot` | formatted_index_{index:03}.png         |
        | `File Should Exist`       | ${OUTPUTDIR}/formatted_index_001.png   |
        """
        if not self.drivers.current:
            self.info('Cannot capture screenshot because no browser is open.')
            return
        path = self._get_screenshot_path(filename)
        self._create_directory(path)
        if not self.driver.save_screenshot(path):
            raise RuntimeError("Failed to save screenshot '{}'.".format(path))
        self._embed_to_log(path, 800)
        return path

    @keyword
    def capture_element_screenshot(self, locator, filename='selenium-element-screenshot-{index}.png'):
        """Captures screenshot from the element identified by ``locator`` and embeds it into log file.

        See `Capture Page Screenshot` for details about ``filename`` argument.
        See the `Locating elements` section for details about the locator
        syntax.

        An absolute path to the created element screenshot is returned.

        Support for capturing the screenshot from a element has limited support
        among browser vendors. Please check the browser vendor driver documentation
        does the browser support capturing a screenshot from a element.

        

        Examples:
        | `Capture Element Screenshot` | id:image_id |                                |
        | `Capture Element Screenshot` | id:image_id | ${OUTPUTDIR}/id_image_id-1.png |
        """
        if not self.drivers.current:
            self.info('Cannot capture screenshot from element because no browser is open.')
            return
        path = self._get_screenshot_path(filename)
        self._create_directory(path)
        element = self.find_element(locator, required=True)
        if not element.screenshot(path):
            raise RuntimeError("Failed to save element screenshot '{}'.".format(path))
        self._embed_to_log(path, 400)
        return path

    def _get_screenshot_path(self, filename):
        directory = self.ctx.screenshot_root_directory or self.log_dir
        filename = filename.replace('/', os.sep)
        index = 0
        while True:
            index += 1
            formatted = filename.format(index=index)
            path = os.path.join(directory, formatted)
            # filename didn't contain {index} or unique path was found
            if formatted == filename or not os.path.exists(path):
                return path

    def _create_directory(self, path):
        target_dir = os.path.dirname(path)
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)

    def _embed_to_log(self, path, width):
        # Image is shown on its own row and thus previous row is closed on
        # purpose. Depending on Robot's log structure is a bit risky.
        self.info('</td></tr><tr><td colspan="3">'
                  '<a href="{src}"><img src="{src}" width="{width}px"></a>'
                  .format(src=get_link_path(path, self.log_dir), width=width), html=True)



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.api import logger
from robot.api.deco import keyword



class Assertion(object):
    """
    Assertion handles all the assertions of rbfDB library.
    """

    @keyword(name=None, tags=("Assertions",))
    def check_if_exists_in_database(self, selectStatement, sansTran=False):
        """
        *Check if any row would be returned by given the input `selectStatement`*. If there are no results, then this will
        throw an AssertionError. 
      
        *Arguments*

        ``selectStatement``: SELECT query 

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples*        
        
        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |

        
        | `Check If Exists In Database` | SELECT id FROM person WHERE first_name = 'Franz Allan' | # PASS |
        | `Check If Exists In Database` | SELECT id FROM person WHERE first_name = 'John' | # FAIL |


        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Check If Exists In Database` | SELECT id FROM person WHERE first_name = 'John' | True |
        """
        logger.info ('Executing : Check If Exists In Database  |  %s ' % selectStatement)
        if not self.query(selectStatement, sansTran):
            raise AssertionError("Expected to have have at least one row from '%s' "
                                 "but got 0 rows." % selectStatement)

    
    @keyword(name=None, tags=("Assertions",))
    def check_if_not_exists_in_database(self, selectStatement, sansTran=False):
        """
        *This is the negation of `check_if_exists_in_database`.*

        Check if no rows would be returned by given the input `selectStatement`. If there are any results, then this
        will throw an AssertionError. 
        
        *Arguments*

        ``selectStatement``: SELECT query 

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples*

        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |

        
        | `Check If Not Exists In Database` | SELECT id FROM person WHERE first_name = 'John' | # PASS |
        | `Check If Not Exists In Database` | SELECT id FROM person WHERE first_name = 'Franz Allan' | # FAIL |


        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Check If Not Exists In Database` | SELECT id FROM person WHERE first_name = 'John' | True |
        """
        logger.info('Executing : Check If Not Exists In Database  |  %s ' % selectStatement)
        queryResults = self.query(selectStatement, sansTran)
        if queryResults:
            raise AssertionError("Expected to have have no rows from '%s' "
                                 "but got some rows : %s." % (selectStatement, queryResults))

    
    @keyword(name=None, tags=("Assertions",))
    def row_count_is_0(self, selectStatement, sansTran=False):
        """
        *Check if any rows are returned from the submitted `selectStatement`.* 
        If there are, then this will throw an AssertionError.
        
        *Arguments*

        ``selectStatement``: SELECT query 

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples* 

        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |

        When you have the following assertions in your robot
        | `Row Count is 0` | SELECT id FROM person WHERE first_name = 'Franz Allan' | # FAIL |
        | `Row Count is 0` | SELECT id FROM person WHERE first_name = 'John' | # PASS |

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Row Count is 0` | SELECT id FROM person WHERE first_name = 'John' | True |
        """
        logger.info('Executing : Row Count Is 0  |  %s ' % selectStatement)
        num_rows = self.row_count(selectStatement, sansTran)
        if num_rows > 0:
            raise AssertionError("Expected zero rows to be returned from '%s' "
                                 "but got rows back. Number of rows returned was %s" % (selectStatement, num_rows))

    
    @keyword(name=None, tags=("Assertions",))
    def row_count_is_equal_to_x(self, selectStatement, numRows, sansTran=False):
        """
        *Check if the number of rows returned from `selectStatement` is equal to the value submitted.* If not, then this
        will throw an AssertionError. 

        *Arguments*

        ``selectStatement``: SELECT query
        
        ``numRows``: Expected number of rows to be returned from `selectStatement`

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples* 
        
        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |
        |  2 | Jerry       | Schneider |

        
        | `Row Count Is Equal To X` | SELECT id FROM person | 1 | # FAIL |
        | `Row Count Is Equal To X` | SELECT id FROM person WHERE first_name = 'John' | 0 | # PASS |
       
        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Row Count Is Equal To X` | SELECT id FROM person WHERE first_name = 'John' | 0 | True |
        """
        logger.info('Executing : Row Count Is Equal To X  |  %s  |  %s ' % (selectStatement, numRows))
        num_rows = self.row_count(selectStatement, sansTran)
        if num_rows != int(numRows.encode('ascii')):
            raise AssertionError("Expected same number of rows to be returned from '%s' "
                                 "than the returned rows of %s" % (selectStatement, num_rows))

    
    @keyword(name=None, tags=("Assertions",))
    def row_count_is_greater_than_x(self, selectStatement, numRows, sansTran=False):
        """
        *Check if the number of rows returned from `selectStatement` is greater than the value submitted.* 
        If not, then this will throw an AssertionError. 

        *Arguments*

        ``selectStatement``: SELECT query
        
        ``numRows``: Expected number of rows to be returned from `selectStatement` 

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples* 
        
        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |
        |  2 | Jerry       | Schneider |

        When you have the following assertions in your robot
        | `Row Count Is Greater Than X` | SELECT id FROM person | 1 | # PASS |
        | `Row Count Is Greater Than X` | SELECT id FROM person WHERE first_name = 'John' | 0 | # FAIL |

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Row Count Is Greater Than X` | SELECT id FROM person | 1 | True |
        """
        logger.info('Executing : Row Count Is Greater Than X  |  %s  |  %s ' % (selectStatement, numRows))
        num_rows = self.row_count(selectStatement, sansTran)
        if num_rows <= int(numRows.encode('ascii')):
            raise AssertionError("Expected more rows to be returned from '%s' "
                                 "than the returned rows of %s" % (selectStatement, num_rows))

    
    @keyword(name=None, tags=("Assertions",))
    def row_count_is_less_than_x(self, selectStatement, numRows, sansTran=False):
        """
        *Check if the number of rows returned from `selectStatement` is less than the value submitted.*
        If not, then this will throw an AssertionError.
        
        *Arguments*

        ``selectStatement``: SELECT query
        
        ``numRows``: Expected number of rows to be returned from `selectStatement` 

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples*  

        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |
        |  2 | Jerry       | Schneider |

        
        | `Row Count Is Less Than X` | SELECT id FROM person | 3 | # PASS |
        | `Row Count Is Less Than X` | SELECT id FROM person WHERE first_name = 'John' | 1 | # FAIL |

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Row Count Is Less Than X` | SELECT id FROM person | 3 | True |
        """
        logger.info('Executing : Row Count Is Less Than X  |  %s  |  %s ' % (selectStatement, numRows))
        num_rows = self.row_count(selectStatement, sansTran)
        if num_rows >= int(numRows.encode('ascii')):
            raise AssertionError("Expected less rows to be returned from '%s' "
                                 "than the returned rows of %s" % (selectStatement, num_rows))

    
    @keyword(name=None, tags=("Assertions",))
    def table_must_exist(self, tableName, sansTran=False):
        """
        *Check if the table given exists in the database.*
        
        *Arguments*

        ``tableName``: Name of the table expected

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples*   

        For example, given we have a table `person` in a database

        | `Table Must Exist` | person | # PASS |
        | `Table Must Exist` | first_name | # FAIL |

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Table Must Exist` | person | True |
        """
        logger.info('Executing : Table Must Exist  |  %s ' % tableName)
        if self.db_api_module_name in ["cx_Oracle"]:
            selectStatement = ("SELECT * FROM all_objects WHERE object_type IN ('TABLE','VIEW') AND owner = SYS_CONTEXT('USERENV', 'SESSION_USER') AND object_name = UPPER('%s')" % tableName)
        elif self.db_api_module_name in ["sqlite3"]:
            selectStatement = ("SELECT name FROM sqlite_master WHERE type='table' AND name='%s' COLLATE NOCASE" % tableName)
        elif self.db_api_module_name in ["ibm_db", "ibm_db_dbi"]:
            selectStatement = ("SELECT name FROM SYSIBM.SYSTABLES WHERE type='T' AND name=UPPER('%s')" % tableName)
        else:
            selectStatement = ("SELECT * FROM information_schema.tables WHERE table_name='%s'" % tableName)
        num_rows = self.row_count(selectStatement, sansTran)
        if num_rows == 0:
            raise AssertionError("Table '%s' does not exist in the db" % tableName)



from robot.api import logger  # noqa: F401 #pylint: disable=unused-import
from TestStack.White import AutomationException, Desktop
from TestStack.White.Configuration import CoreAppXmlConfiguration
from TestStack.White.Factory import InitializeOption  # noqa: E402
from TestStack.White.UIItems.Finders import SearchCriteria  # noqa: E402
from TestStack.White.UIItems.WindowItems import DisplayState, Window  # noqa: F401
from rbfDesktop.keywords.librarycomponent import LibraryComponent
from rbfDesktop.keywords.robotlibcore import keyword

WINDOW_STRATEGIES = {"id": "ByAutomationId", "class_name": "ByClassName"}


class WindowKeywords(LibraryComponent):
    @keyword
    def attach_window(self, locator):
        """Attaches rbfDesktop to a window.

        ``locator`` is the locator of the window or a window object.

        === Window locator syntax ===
        Syntax for a window locator is ``prefix:value``.

        When the locator syntax is used, the window is searched from the currently attached application.
        See `Attach Application By Id`, `Attach Application By Name` or
        `Launch application` for more details about attaching an application.

        The following window locators are available:
        | = Prefix =           | = Description =         |
        | title (or no prefix) | Search by the exact window title. If no prefix is given, the window is searched by title by default. |
        | id                   | Search by AutomationID. |
        | class_name           | Search by class name.   |

        Examples:
        | Attach Window | Main Window                   | # attach window with title         |
        | Attach Window | id:mainWindow                 | # attach window with Automation ID |
        | Attach Window | class_name:NavigationWindow   | # attach window with ClassName     |

        === Window objects ===
        A window can also be attached by directly passing the window object as the ``locator`` parameter value.
        This may be useful if the correct window cannot be found by using the window locator syntax.

        Example:
        | @{windows} | `Get Application Windows` | |
        | Attach Window | ${windows[1]} | # attach window at index 1 in window list |

        When using a window object as the ``locator`` parameter value, the window is attached even if it does not
        belong to the currently attached application.
        Note that when attaching a window that belongs to a different application than the currently attached one,
        attaching the window does not affect what application is attached to the library.

        """
        self.state.window = self._get_window(locator)

    @keyword
    def close_window(self, locator=None):
        """Closes a window.

        ``locator`` is the locator of the window or a window object (optional).

        If no ``locator`` value is given, the currently attached window is closed.
        See `Attach Window` for details about window locators and attaching a window.
        """
        if locator is not None:
            window = self._get_window(locator)
            window.Close()
        else:
            self.state.window.Close()

    @keyword
    def get_application_windows(self):
        """Returns a list of windows belonging to the currently attached application.

        Assumes that an application is attached.
        See `Attach Application By Name` and `Attach Application By Id` for details.
        """
        return list(self.state.app.GetWindows())

    @staticmethod
    @keyword
    def get_desktop_windows():
        """Returns a list of windows on the desktop."""
        return list(Desktop.Instance.Windows())

    @keyword
    def get_window_title(self):
        """Returns title of the currently attached window.

        Assumes that a window is attached. See `Attach Window` for details.
        """
        return self.state.window.Title

    @keyword
    def select_modal_window(self, window_title):
        """Attaches a modal window.

        ``window_title`` is the title of the window.
        """
        self.state.window = self.state.window.ModalWindow(window_title)

    @keyword
    def window_title_should_be(self, expected):
        """Verifies that the title of the currently attached window is ``expected``.

        Assumes that a window is attached. See `Attach Window` for details.
        """
        self.state._verify_string_value(expected, self.state.window.Title)

    @keyword
    def window_title_should_contain(self, expected):
        """Verifies that the title of the currently attached window contains text ``expected``.

        Assumes that a window is attached. See `Attach Window` for details.
        """
        self.state._contains_string_value(expected, self.state.window.Title)

    @keyword
    def maximize_window(self, locator=None):
        """Maximizes a window.

        ``locator`` is the locator of the window or a window object (optional).

        If no ``locator`` value is given, the currently attached window is maximized.
        See `Attach Window` for details about attaching a window and window locator syntax.
        """
        if locator is not None:
            window = self._get_window(locator)
        else:
            window = self.state.window
        window.DisplayState = DisplayState.Maximized

    @keyword
    def window_should_be_maximized(self, locator=None):
        """Verifies that a window is maximized.

        ``locator`` is the locator of the window or a window object (optional).

        If no ``locator`` value is given, the status of the currently attached window is verified.
        See `Attach Window` for details about attaching a window and window locator syntax.
        """
        if locator is not None:
            window = self._get_window(locator)
        else:
            window = self.state.window
        if window.DisplayState != DisplayState.Maximized:
            raise AssertionError("Expected window state to be maximized, but found {}".format(str(window.DisplayState)))

    @keyword
    def minimize_window(self, locator=None):
        """Minimizes a window.

        ``locator`` is the locator of the window or a window object (optional).

        If no ``locator`` value is given, the currently attached window is minimized.
        See `Attach Window` for details about attaching a window and window locator syntax.
        """
        if locator is not None:
            window = self._get_window(locator)
        else:
            window = self.state.window
        window.DisplayState = DisplayState.Minimized

    @keyword
    def window_should_be_minimized(self, locator=None):
        """Verifies that a window is minimized.

        ``locator`` is the locator of the window or a window object (optional).

        If no ``locator`` value is given, the status of the currently attached window is verified.
        See `Attach Window` for details about attaching a window and window locator syntax.
        """
        if locator is not None:
            window = self._get_window(locator)
        else:
            window = self.state.window
        if window.DisplayState != DisplayState.Minimized:
            raise AssertionError("Expected window state to be minimized, but found {}".format(str(window.DisplayState)))

    @keyword
    def restore_window(self, window_title=None):
        """Restores a window.

        ``locator`` is the locator of the window or a window object (optional).

        If no ``locator`` value is given, the currently attached window is restored.
        See `Attach Window` for details about attaching a window and window locator syntax.
        """
        if window_title is not None:
            window = self._get_window(window_title)
        else:
            window = self.state.window
        window.DisplayState = DisplayState.Restored

    @keyword
    def window_should_be_restored(self, window_title=None):
        """Verifies that a window is restored.

        ``locator`` is the locator of the window or a window object (optional).

        If title is not given, currently attached window status is queried.
        See `Attach Window` for more details.
        """
        if window_title is not None:
            window = self._get_window(window_title)
        else:
            window = self.state.window
        if window.DisplayState != DisplayState.Restored:
            raise AssertionError("Expected window state to be restored, but found {}".format(str(window.DisplayState)))

    def _get_window(self, locator):
        if isinstance(locator, Window):
            return locator
        return self._get_window_by_locator(locator)

    def _get_window_by_locator(self, locator):
        search_strategy, locator_value = self._parse_window_locator(locator)
        try:
            if search_strategy == "title":
                return self.state.app.GetWindow(locator_value)
            search_criteria = getattr(SearchCriteria, WINDOW_STRATEGIES[search_strategy])(locator_value)
            return self.state.app.GetWindow(search_criteria, InitializeOption.NoCache)
        except KeyError:
            raise ValueError("'{}' is not a valid locator prefix for a window".format(search_strategy))
        except AttributeError as error_msg:
            error_msg = str(error_msg)
            if "NoneType" in error_msg:
                error_msg = "No application attached."
            raise AttributeError(error_msg)
        except AutomationException as error_msg:
            error_msg = str(error_msg)
            replaced_text = "after waiting for {0} seconds".format(
                CoreAppXmlConfiguration.Instance.FindWindowTimeout / 1000.0
            )
            raise AutomationException(error_msg.replace("after waiting for 30 seconds", replaced_text), "")

    @staticmethod
    def _parse_window_locator(locator):
        if ":" not in locator:
            locator = "title:" + locator
        idx = locator.index(":")
        return locator[:idx], locator[(idx + 1):]



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from robot.api import logger
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError

from rbfWEB.utils import is_noney, timestr_to_secs

from .context import ContextAware
from .robotlibcore import PY2


class LibraryComponent(ContextAware):

    def info(self, msg, html=False):
        logger.info(msg, html)

    def debug(self, msg, html=False):
        logger.debug(msg, html)

    def log(self, msg, level='INFO', html=False):
        if not is_noney(level):
            logger.write(msg, level.upper(), html)

    def warn(self, msg, html=False):
        logger.warn(msg, html)

    def log_source(self, loglevel='INFO'):
        self.ctx.log_source(loglevel)

    def assert_page_contains(self, locator, tag=None, message=None,
                             loglevel='TRACE'):
        if not self.find_element(locator, tag, required=False):
            self.log_source(loglevel)
            if is_noney(message):
                message = ("Page should have contained %s '%s' but did not."
                           % (tag or 'element', locator))
            raise AssertionError(message)
        logger.info("Current page contains %s '%s'."
                    % (tag or 'element', locator))

    def assert_page_not_contains(self, locator, tag=None, message=None,
                                 loglevel='TRACE'):
        if self.find_element(locator, tag, required=False):
            self.log_source(loglevel)
            if is_noney(message):
                message = ("Page should not have contained %s '%s'."
                           % (tag or 'element', locator))
            raise AssertionError(message)
        logger.info("Current page does not contain %s '%s'."
                    % (tag or 'element', locator))

    def get_timeout(self, timeout=None):
        if is_noney(timeout):
            return self.ctx.timeout
        return timestr_to_secs(timeout)

    @property
    def log_dir(self):
        try:
            logfile = BuiltIn().get_variable_value('${LOG FILE}')
            if logfile == 'NONE':
                return BuiltIn().get_variable_value('${OUTPUTDIR}')
            return os.path.dirname(logfile)
        except RobotNotRunningError:
            return os.getcwdu() if PY2 else os.getcwd()



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rbfWEB.base import LibraryComponent, keyword


class TableElementKeywords(LibraryComponent):

    @keyword
    def get_table_cell(self, locator, row, column, loglevel='TRACE'):
        """Returns contents of table cell.

        The table is located using the ``locator`` argument and its cell
        found using ``row`` and ``column``. See the `Locating elements`
        section for details about the locator syntax.

        Both row and column indexes start from 1, and header and footer
        rows are included in the count. It is possible to refer to rows
        and columns from the end by using negative indexes so that -1
        is the last row/column, -2 is the second last, and so on.

        All ``<th>`` and ``<td>`` elements anywhere in the table are
        considered to be cells.

        See `Page Should Contain` for explanation about the ``loglevel``
        argument.
        """
        row = int(row)
        column = int(column)
        if row == 0 or column == 0:
            raise ValueError('Both row and column must be non-zero, '
                             'got row %d and column %d.' % (row, column))
        try:
            cell = self._get_cell(locator, row, column)
        except AssertionError:
            self.log_source(loglevel)
            raise
        return cell.text

    def _get_cell(self, locator, row, column):
        rows = self._get_rows(locator, row)
        if len(rows) < abs(row):
            raise AssertionError("Table '%s' should have had at least %d "
                                 "rows but had only %d."
                                 % (locator, abs(row), len(rows)))
        index = row - 1 if row > 0 else row
        cells = rows[index].find_elements_by_xpath('./th|./td')
        if len(cells) < abs(column):
            raise AssertionError("Table '%s' row %d should have had at "
                                 "least %d columns but had only %d."
                                 % (locator, row, abs(column), len(cells)))
        index = column - 1 if column > 0 else column
        return cells[index]

    def _get_rows(self, locator, count):
        # Get rows in same order as browsers render them.
        table = self.find_element(locator, tag='table')
        rows = table.find_elements_by_xpath("./thead/tr")
        if count < 0 or len(rows) < count:
            rows.extend(table.find_elements_by_xpath("./tbody/tr"))
        if count < 0 or len(rows) < count:
            rows.extend(table.find_elements_by_xpath("./tfoot/tr"))
        return rows

    @keyword
    def table_cell_should_contain(self, locator, row, column, expected, loglevel='TRACE'):
        """Verifies table cell contains text ``expected``.

        See `Get Table Cell` that this keyword uses internally for
        explanation about accepted arguments.
        """
        content = self.get_table_cell(locator, row, column, loglevel)
        if expected not in content:
            self.ctx.log_source(loglevel)
            raise AssertionError("Table '%s' cell on row %s and column %s "
                                 "should have contained text '%s' but it had "
                                 "'%s'."
                                 % (locator, row, column, expected, content))
        self.info("Table cell contains '%s'." % content)

    @keyword
    def table_column_should_contain(self, locator, column, expected, loglevel='TRACE'):
        """Verifies table column contains text ``expected``.

        The table is located using the ``locator`` argument and its column
        found using ``column``. See the `Locating elements` section for
        details about the locator syntax.

        Column indexes start from 1. It is possible to refer to columns
        from the end by using negative indexes so that -1 is the last column,
        -2 is the second last, and so on.

        If a table contains cells that span multiple columns, those merged
        cells count as a single column.

        See `Page Should Contain Element` for explanation about the
        ``loglevel`` argument.
        """
        element = self._find_by_column(locator, column, expected)
        if element is None:
            self.ctx.log_source(loglevel)
            raise AssertionError("Table '%s' column %s did not contain text "
                                 "'%s'." % (locator, column, expected))

    @keyword
    def table_footer_should_contain(self, locator, expected, loglevel='TRACE'):
        """Verifies table footer contains text ``expected``.

        Any ``<td>`` element inside ``<tfoot>`` element is considered to
        be part of the footer.

        The table is located using the ``locator`` argument. See the
        `Locating elements` section for details about the locator syntax.

        See `Page Should Contain Element` for explanation about the
        ``loglevel`` argument.
        """
        element = self._find_by_footer(locator, expected)
        if element is None:
            self.ctx.log_source(loglevel)
            raise AssertionError("Table '%s' footer did not contain text "
                                 "'%s'." % (locator, expected))

    @keyword
    def table_header_should_contain(self, locator, expected, loglevel='TRACE'):
        """Verifies table header contains text ``expected``.

        Any ``<th>`` element anywhere in the table is considered to be
        part of the header.

        The table is located using the ``locator`` argument. See the
        `Locating elements` section for details about the locator syntax.

        See `Page Should Contain Element` for explanation about the
        ``loglevel`` argument.
        """
        element = self._find_by_header(locator, expected)
        if element is None:
            self.ctx.log_source(loglevel)
            raise AssertionError("Table '%s' header did not contain text "
                                 "'%s'." % (locator, expected))

    @keyword
    def table_row_should_contain(self, locator, row, expected, loglevel='TRACE'):
        """Verifies that table row contains text ``expected``.

        The table is located using the ``locator`` argument and its column
        found using ``column``. See the `Locating elements` section for
        details about the locator syntax.

        Row indexes start from 1. It is possible to refer to rows
        from the end by using negative indexes so that -1 is the last row,
        -2 is the second last, and so on.

        If a table contains cells that span multiple rows, a match
        only occurs for the uppermost row of those merged cells.

        See `Page Should Contain Element` for explanation about the
        ``loglevel`` argument.
        """
        element = self._find_by_row(locator, row, expected)
        if element is None:
            self.ctx.log_source(loglevel)
            raise AssertionError("Table '%s' row %s did not contain text "
                                 "'%s'." % (locator, row, expected))

    @keyword
    def table_should_contain(self, locator, expected, loglevel='TRACE'):
        """Verifies table contains text ``expected``.

        The table is located using the ``locator`` argument. See the
        `Locating elements` section for details about the locator syntax.

        See `Page Should Contain Element` for explanation about the
        ``loglevel`` argument.
        """
        element = self._find_by_content(locator, expected)
        if element is None:
            self.ctx.log_source(loglevel)
            raise AssertionError("Table '%s' did not contain text '%s'."
                                 % (locator, expected))

    def _find_by_content(self, table_locator, content):
        return self._find(table_locator, '//*', content)

    def _find_by_header(self, table_locator, content):
        return self._find(table_locator, '//th', content)

    def _find_by_footer(self, table_locator, content):
        return self._find(table_locator, '//tfoot//td', content)

    def _find_by_row(self, table_locator, row, content):
        position = self._index_to_position(row)
        locator = '//tr[{}]'.format(position)
        return self._find(table_locator, locator, content)

    def _find_by_column(self, table_locator, col, content):
        position = self._index_to_position(col)
        locator = '//tr//*[self::td or self::th][{}]'.format(position)
        return self._find(table_locator, locator, content)

    def _index_to_position(self, index):
        index = int(index)
        if index == 0:
            raise ValueError('Row and column indexes must be non-zero.')
        if index > 0:
            return str(index)
        if index == -1:
            return 'position()=last()'
        return 'position()=last()-{}'.format(abs(index) - 1)

    def _find(self, table_locator, locator, content):
        table = self.find_element(table_locator)
        elements = self.find_elements(locator, parent=table)
        for element in elements:
            if content is None:
                return element
            if element.text and content in element.text:
                return element
        return None



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import six
from rbfExcel.base import ExcelLibrary
from rbfExcel.version import __version__
from rbfExcel.utils import DateFormat, NumberFormat, BoolFormat




class rbfExcel(ExcelLibrary):
    """
    rbfExcel library provides some keywords to allow opening, reading, writing, and saving Excel files from Robot Framework.

    *Before running tests*

    Prior to running tests, rbfExcel must first be imported into your Robot test suite.

    Example:
        | Library | rbfExcel |

    To setup some Excel configurations related to date format and number format, use these arguments

        *Excel Date Time format*
        | Date Format       | Default: `yyyy-mm-dd`         |
        | Time Format       | Default: `HH:MM:SS AM/PM`     |
        | Date Time Format  | Default: `yyyy-mm-dd HH:MM`   |
        For more information, check this article
        https://support.office.com/en-us/article/format-numbers-as-dates-or-times-418bd3fe-0577-47c8-8caa-b4d30c528309

        *Excel Number format*
        | Decimal Separator     | Default: `.`  |
        | Thousand Separator    | Default: `,`  |
        | Precision             | Default: `2`  |

        *Excel Boolean format*
        | Boolean Format        | Default: `Yes/No`  |

    Example:
        | Library | rbfExcel | date_format='dd/mm/yyyy'
    """
    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = __version__

    def __init__(self,
                 date_format='yyyy-mm-dd', time_format='HH:MM:SS AM/PM', datetime_format='yyyy-mm-dd HH:MM',
                 decimal_sep='.', thousand_sep=',', precision='2', bool_format='Yes/No'):
        logging.basicConfig()
        logging.getLogger().setLevel(logging.INFO)
        logger = logging.getLogger(__name__)
        logger.info('rbfExcel::Robotframework Excel Library')
        super(rbfExcel, self).__init__(
            DateFormat(date_format, time_format, datetime_format),
            NumberFormat(decimal_sep, thousand_sep, precision),
            BoolFormat(bool_format))



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from robot.api import logger
from robot.libraries.BuiltIn import BuiltIn

from .Notify import Notify
from .version import __version__



class rbfSlack(Notify):
    """rbfSlack is a slack module for Robot Framework.

    This library provides functionality within Robot Framework to push automated test status to SLack Channel.

To utilize this, pass rbfSlack as robot listener argument during execution. 
For example: robot --listener rbfSlack.Notify:${slackChannelID}:${slackToken}:${testSuite}:${builtBy}:${builtWhy}

Values of Parameters ${slackChannelID}, ${slackToken}, ${testSuite} are configured and picked up at run-time from global.resource file.
Values of Parameters ${builtBy} and ${builtWhy} are collected from the Build tool (Bamboo/Jenkins)
    
    """
    ROBOT_LIBRARY_SCOPE = 'TEST SUITE'
    ROBOT_LIBRARY_VERSION = __version__



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from .connection_manager import ConnectionManager
from .query import Query
from .assertion import Assertion
from .version import __version__



class rbfDB(ConnectionManager, Query, Assertion):
    """
    rbfDB library contains utilities meant for rbf usage.

    This can allow you to query your database after an action has been made to verify the results.

    This is `compatible*` with any Database API Specification 2.0 module.



    References:

     + Database API Specification 2.0 - http://www.python.org/dev/peps/pep-0249/

     + Lists of DB API 2.0 - http://wiki.python.org/moin/DatabaseInterfaces

     + Python Database Programming - http://wiki.python.org/moin/DatabaseProgramming/

    Notes:



    `compatible* - or at least theoretically it should be compatible. Currently tested only with postgresql
    (using psycopg2).`

    Example Usage:
    | # Setup |
    | Connect to Database |
    | # Guard assertion (verify that test started in expected state). |
    | Check if not exists in database | select id from person where first_name = 'Franz Allan' and last_name = 'See' |
    | # Drive UI to do some action |
    | Go To | http://localhost/person/form.html | | # From selenium library |
    | Input Text |  name=first_name | Franz Allan | # From selenium library |
    | Input Text |  name=last_name | See | # From selenium library |
    | Click Button | Save | | # From selenium library |
    | # Log results |
    | @{queryResults} | Query | select * from person |
    | Log Many | @{queryResults} |
    | # Verify if persisted in the database |
    | Check if exists in database | select id from person where first_name = 'Franz Allan' and last_name = 'See' |
    | # Teardown |
    | Disconnect from Database |
    """

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = __version__



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rbfWEB.base import LibraryComponent, keyword


class FrameKeywords(LibraryComponent):

    @keyword
    def select_frame(self, locator):
        """Sets frame identified by ``locator`` as the current frame.

        See the `Locating elements` section for details about the locator
        syntax.

        Works both with frames and iframes. Use `Unselect Frame` to cancel
        the frame selection and return to the main frame.

        Example:
        | `Select Frame`   | top-frame | # Select frame with id or name 'top-frame'   |
        | `Click Link`     | example   | # Click link 'example' in the selected frame |
        | `Unselect Frame` |           | # Back to main frame.                        |
        | `Select Frame`   | //iframe[@name='xxx'] | # Select frame using xpath       |
        """
        self.info("Selecting frame '%s'." % locator)
        element = self.find_element(locator)
        self.driver.switch_to.frame(element)

    @keyword
    def unselect_frame(self):
        """Sets the main frame as the current frame.

        In practice cancels the previous `Select Frame` call.
        """
        self.driver.switch_to.default_content()

    @keyword
    def current_frame_should_contain(self, text, loglevel='TRACE'):
        """Verifies that current frame contains ``text``.

        See `Page Should Contain` for explanation about the ``loglevel``
        argument.

        
        """
        if not self.is_text_present(text):
            self.log_source(loglevel)
            raise AssertionError("Frame should have contained text '%s' "
                                 "but did not." % text)
        self.info("Current frame contains text '%s'." % text)

    

    @keyword
    def current_frame_should_not_contain(self, text, loglevel='TRACE'):
        """Verifies that current frame does not contains ``text``.

        See `Page Should Contain` for explanation about the ``loglevel``
        argument.
        """
        if self.is_text_present(text):
            self.log_source(loglevel)
            raise AssertionError("Frame should not have contained text '%s' "
                                 "but it did." % text)
        self.info("Current frame did not contain text '%s'." % text)

    @keyword
    def frame_should_contain(self, locator, text, loglevel='TRACE'):
        """Verifies that frame identified by ``locator`` contains ``text``.

        See the `Locating elements` section for details about the locator
        syntax.

        See `Page Should Contain` for explanation about the ``loglevel``
        argument.
        """
        if not self._frame_contains(locator, text):
            self.log_source(loglevel)
            raise AssertionError("Frame '%s' should have contained text '%s' "
                                 "but did not." % (locator, text))
        self.info("Frame '%s' contains text '%s'." % (locator, text))

    def _frame_contains(self, locator, text):
        element = self.find_element(locator)
        self.driver.switch_to.frame(element)
        self.info("Searching for text from frame '%s'." % locator)
        found = self.is_text_present(text)
        self.driver.switch_to.default_content()
        return found



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import threading

try:
    import cv2
    import numpy as np
except ImportError:
    raise ImportError('Importing cv2 failed. Make sure you have opencv-python installed.')

from mss import mss
from PIL import Image
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures.thread import _threads_queues
from functools import wraps
from robot.api import logger
from robot.utils import get_link_path, abspath, timestr_to_secs, is_truthy
from robot.libraries.BuiltIn import BuiltIn
from .pygtk import _take_gtk_screenshot, _take_partial_gtk_screenshot, _take_gtk_screen_size, _grab_gtk_pb, _record_gtk
from .utils import _norm_path, _compression_value_conversion, _pil_quality_conversion, suppress_stderr

_THREAD_POOL = ThreadPoolExecutor()


def run_in_background(f, executor=None):
    @wraps(f)
    def wrap(*args, **kwargs):
        return (executor or _THREAD_POOL).submit(f, *args, **kwargs)
    return wrap


class Client:

    def __init__(self, screenshot_module=None, screenshot_directory=None, format='png', quality=50, delay=0, fps=8):
        self._screenshot_module = screenshot_module
        self._given_screenshot_dir = _norm_path(screenshot_directory)
        self._format = format
        self._quality = quality
        self._delay = delay
        self.frames = []
        self.name = 'screenshot'
        self.path = None
        self.embed = False
        self.embed_width = None
        self.fps = fps
        self._stop_condition = threading.Event()
        self.futures = None

    @property
    def _screenshot_dir(self):
        return self._given_screenshot_dir or self._log_dir

    @property
    def _log_dir(self):
        variables = BuiltIn().get_variables()
        outdir = variables['${OUTPUTDIR}']
        log = variables['${LOGFILE}']
        log = os.path.dirname(log) if log != 'NONE' else '.'
        return _norm_path(os.path.join(outdir, log))

    def set_screenshot_directory(self, path):
        path = _norm_path(path)
        if not os.path.isdir(path):
            raise RuntimeError("Directory '%s' does not exist." % path)
        old = self._screenshot_dir
        self._given_screenshot_dir = path
        return old

    def _get_screenshot_path(self, basename, format, directory):
        directory = _norm_path(directory) if directory else self._screenshot_dir
        if basename.lower().endswith(('.jpg', '.jpeg', '.png', '.webp', '.webm')):
            return os.path.join(directory, basename)
        index = 0
        while True:
            index += 1
            path = os.path.join(directory, "%s_%d.%s" % (basename, index, format))
            if not os.path.exists(path):
                return path

    @staticmethod
    def _validate_screenshot_path(path):
        path = abspath(_norm_path(path))
        if not os.path.exists(os.path.dirname(path)):
            raise RuntimeError("Directory '%s' where to save the screenshot "
                               "does not exist" % os.path.dirname(path))
        return path

    def _save_screenshot_path(self, basename, format):
        path = self._get_screenshot_path(basename, format, self._screenshot_dir)
        return self._validate_screenshot_path(path)

    def take_screenshot(self, name, format, quality, width='800px', delay=0):
        delay = delay or self._delay
        if delay:
            time.sleep(timestr_to_secs(delay))
        path = self._take_screenshot_client(name, format, quality)
        self._embed_screenshot(path, width)
        return path

    def _take_screenshot_client(self, name, format, quality):
        format = (format or self._format).lower()
        quality = quality or self._quality
        if self._screenshot_module and self._screenshot_module.lower() == 'pygtk':
            return self._take_screenshot_client_gtk(name, format, quality)
        else:
            return self._take_screenshot_client_mss(name, format, quality)

    def _take_screenshot_client_gtk(self, name, format, quality):
        format = 'jpeg' if format == 'jpg' else format
        if format == 'png':
            quality = _compression_value_conversion(quality)
        path = self._save_screenshot_path(name, format)
        if format == 'webp':
            png_img = _take_gtk_screenshot(path, 'png', _compression_value_conversion(100))
            im = Image.open(png_img)
            im.save(path, format, quality=quality)
            return path
        return _take_gtk_screenshot(path, format, quality)

    def _take_screenshot_client_mss(self, name, format, quality):
        if format in ['jpg', 'jpeg', 'webp']:
            with mss() as sct:
                sct_img = sct.grab(sct.monitors[0])
                img = Image.frombytes('RGB', sct_img.size, sct_img.bgra, 'raw', 'BGRX')
                path = self._save_screenshot_path(name, format)
                img.save(path, quality=quality if format == 'webp' else _pil_quality_conversion(quality))
            return path
        elif format == 'png':
            with mss() as sct:
                path_name = self._save_screenshot_path(name, format)
                sct.compression_level = _compression_value_conversion(quality)
                path = sct.shot(mon=-1, output='%s' % path_name)
            return path
        else:
            raise RuntimeError("Invalid screenshot format.")

    def take_multiple_screenshots(self, name, format, quality, screenshot_number, delay_time):
        self.frames = []
        quality = quality or self._quality
        format = (format or self._format).lower()
        format = 'jpeg' if format == 'jpg' else format
        if format == 'png':
            quality = _compression_value_conversion(quality)
        elif format == 'jpeg':
            quality = _pil_quality_conversion(quality)
        delay_time = timestr_to_secs(delay_time)
        self.grab_frames(name, format, quality, delay=delay_time, shot_number=int(screenshot_number))

    def take_partial_screenshot(self, name, format, quality,
                                left, top, width, height, embed, embed_width):
        left = int(left)
        top = int(top)
        width = int(width)
        height = int(height)
        format = (format or self._format).lower()
        quality = quality or self._quality

        if self._screenshot_module and self._screenshot_module.lower() == 'pygtk':
            format = 'jpeg' if format == 'jpg' else format
            if format == 'png':
                quality = _compression_value_conversion(quality)
            path = self._save_screenshot_path(name, format)
            path = _take_partial_gtk_screenshot(path, format, quality, left, top, width, height)
        else:
            try:
                original_image = self.take_screenshot(name, format, quality)
                image = Image.open(original_image)
                box = (left, top, width, height)
                cropped_image = image.crop(box)
                os.remove(original_image)
                path = self._save_screenshot_path(basename=name, format=format)
                cropped_image.save(path, format)
            except IOError:
                raise IOError('File not found.')
            except RuntimeError:
                raise RuntimeError('Taking screenshot failed.')
            except SystemError:
                raise SystemError("Top and left parameters must be lower than screen resolution.")
        if is_truthy(embed):
            self._embed_screenshot(path, embed_width)
        return path

    def take_screenshot_without_embedding(self, name, format, quality, delay):
        delay = delay or self._delay
        if delay:
            time.sleep(timestr_to_secs(delay))
        path = self._take_screenshot_client(name, format, quality)
        self._link_screenshot(path)
        return path

    def start_gif_recording(self, name, size_percentage,
                            embed, embed_width):
        self.name = name
        self.embed = embed
        self.embed_width = embed_width
        self.futures = self.grab_frames(name, size_percentage=size_percentage)

    def _close_threads(self):
        if self.futures._exception:
            raise self.futures._exception
        _THREAD_POOL._threads.clear()
        _threads_queues.clear()

    def stop_gif_recording(self):
        self._close_threads()
        path = self._save_screenshot_path(basename=self.name, format='gif')
        self.frames[0].save(path, save_all=True, append_images=self.frames[1:],
                            duration=125, optimize=True, loop=0)
        if is_truthy(self.embed):
            self._embed_screenshot(path, self.embed_width)
        self.frames = []
        return path

    @run_in_background
    def grab_frames(self, name, format=None, quality=None, size_percentage=0.5, delay=0, shot_number=None):
        if self._screenshot_module and self._screenshot_module.lower() == 'pygtk':
            self._grab_frames_gtk(size_percentage, delay, shot_number)
        else:
            self._grab_frames_mss(size_percentage, delay, shot_number)
        if shot_number:
            for img in self.frames:
                path = self._save_screenshot_path(basename=name, format=format)
                img.save(path, format=format, quality=quality, compress_level=quality)

    def _grab_frames_gtk(self, size_percentage, delay, shot_number):
        width, height = _take_gtk_screen_size()
        w = int(width * size_percentage)
        h = int(height * size_percentage)
        while True:
            pb = _grab_gtk_pb()
            img = Image.frombuffer('RGB', (width, height), pb.get_pixels(), 'raw', 'RGB').resize((w, h))
            self.frames.append(img)
            if delay:
                time.sleep(timestr_to_secs(delay))
            if shot_number and len(self.frames) == int(shot_number):
                break
            time.sleep(0.125)

    def _grab_frames_mss(self, size_percentage, delay, shot_number):
        with mss() as sct:
            width = int(sct.grab(sct.monitors[0]).size.width * size_percentage)
            height = int(sct.grab(sct.monitors[0]).size.height * size_percentage)
            while True:
                sct_img = sct.grab(sct.monitors[0])
                img = Image.frombytes('RGB', sct_img.size, sct_img.bgra, 'raw', 'BGRX').resize((width, height))
                self.frames.append(img)
                if delay:
                    time.sleep(timestr_to_secs(delay))
                if shot_number and len(self.frames) == int(shot_number):
                    break

    def _embed_screenshot(self, path, width):
        link = get_link_path(path, self._log_dir)
        logger.info('<a href="%s"><img src="%s" width="%s"></a>' % (link, link, width), html=True)

    def _embed_video(self, path, width):
        link = get_link_path(path, self._log_dir)
        logger.info('<a href="%s"><video width="%s" autoplay><source src="%s" type="video/webm"></video></a>' %
                    (link, width, link), html=True)

    def _link_screenshot(self, path):
        link = get_link_path(path, self._log_dir)
        logger.info("Screenshot saved to '<a href=\"%s\">%s</a>'." % (link, path), html=True)

    def start_video_recording(self, name, fps, embed, embed_width):
        self._stop_condition.clear()
        self.name = name
        try:
            self.fps = int(fps)
        except ValueError:
            raise ValueError('The fps argument must be of type integer.')
        self.embed = embed
        self.embed_width = embed_width
        self.path = self._save_screenshot_path(basename=self.name, format='webm')
        self.futures = self.capture_screen(self.path, self.fps)

    def stop_video_recording(self):
        self._stop_condition.set()
        self._close_threads()
        if is_truthy(self.embed):
            self._embed_video(self.path, self.embed_width)
        return self.path

    @run_in_background
    def capture_screen(self, path, fps):
        if self._screenshot_module and self._screenshot_module.lower() == 'pygtk':
            _record_gtk(path, fps, stop=self._stop_condition)
        else:
            self._record_mss(path, fps)

    def _record_mss(self, path, fps):
        fourcc = cv2.VideoWriter_fourcc(*'VP08')
        with mss() as sct:
            sct_img = sct.grab(sct.monitors[1])
        width = int(sct_img.width)
        height = int(sct_img.height)
        with suppress_stderr():
            vid = cv2.VideoWriter('%s' % path, fourcc, fps, (width, height))
        while not self._stop_condition.isSet():
            with mss() as sct:
                sct_img = sct.grab(sct.monitors[1])
            numpy_array = np.array(sct_img)
            frame = cv2.cvtColor(numpy_array, cv2.COLOR_RGBA2RGB)
            vid.write(frame)
        vid.release()
        cv2.destroyAllWindows()



#!/usr/bin/env python

import re
from os.path import abspath, dirname, join
from setuptools import setup, find_packages


CURDIR = dirname(abspath(__file__))

CLASSIFIERS = '''
Development Status :: 5 - Production/Stable
License :: OSI Approved :: Apache Software License
Operating System :: OS Independent
Programming Language :: Python :: 3
Topic :: Software Development :: Testing
Framework :: Robot Framework :: Library
'''.strip().splitlines()
with open(join(CURDIR, 'version.py'), encoding="utf-8") as f:
    VERSION = re.search("__version__ = '(.*)'", f.read()).group(1)
with open(join(CURDIR, 'requirements.txt')) as f:
    REQUIREMENTS = f.read().splitlines()

setup(
    name='rbf',
    version=VERSION,
    description="Robot Framework test library for automation of Web, Mobile, REST APIs, SOAP Services, zOS, FTP, DB, Desktop Applications",
    author="Rakesh Ummadisetty",
    author_email='rakesh.ummadisetty@dnb.no',
    url='https://AB62031@develop.dnb.no/git/scm/ab/4748-dnb_rbf-library.git',
    license          = 'Apache License 2.0',
    keywords         = 'robotframework testing testautomation',
    platforms        = 'any',
    install_requires = REQUIREMENTS,  
    classifiers=CLASSIFIERS,
    packages=['rbfFTP', 'rbfZOS', 'rbfREST', 'rflint', 'rflint.rules', 'rflint.parser', 'rbfWEB', 'rbfWEB.base', 'rbfWEB.keywords', 'rbfWEB.locators', 'rbfWEB.utils', 'rbfWEB.utils.events', 'rbfDB', 'rbfDesktop', 'rbfDesktop.keywords', 'rbfDesktop.keywords.items', 'rbfDesktop.utils', 'rbfDataDriver', 'rbfMetrics', 'rbfHTTP', 'rbfExcel', 'rbfSlack', 'rbfVideoRecorder', 'rbfMobile'],
    package_data={"rbfDesktop": ["bin/*.dll"]},
    include_package_data=True,
    entry_points={
          'console_scripts': [
              'robotmetrics=rbfMetrics.runner:main',
          ]
      }     
)



from TestStack.White.UIItems.TabItems import Tab
from rbfDesktop.keywords.librarycomponent import LibraryComponent
from rbfDesktop.keywords.robotlibcore import keyword


class TabKeywords(LibraryComponent):
    @keyword
    def select_tab_page(self, locator, title):
        """
        Selects a tab page.

        ``locator`` is the locator of the tab control item or Tab item object.
        Locator syntax is explained in `Item locators`.

        ``title`` is the title of the tab page.
        """
        tab_control = self.state._get_typed_item_by_locator(Tab, locator)
        tab_control.SelectTabPage(title)

    @keyword
    def select_tab_page_by_index(self, locator, index):
        """
        Selects a tab page by index.

        ``locator`` is the locator of the tab control item or Tab item object.
        Locator syntax is explained in `Item locators`.

        ``index`` is integer index the tab page. Indexing starts from 0.
        """
        tab_control = self.state._get_typed_item_by_locator(Tab, locator)
        tab_control.SelectTabPage(int(index))

    @keyword
    def get_tab_pages(self, locator):
        """
        Gets all tab pages and returns them as a list.

        ``locator`` is the locator of the tab control item or Tab item object.
        Locator syntax is explained in `Item locators`.
        """
        tab_control = self.state._get_typed_item_by_locator(Tab, locator)
        return tab_control.Pages



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from selenium.common.exceptions import WebDriverException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from rbfWEB.base import keyword, LibraryComponent
from rbfWEB.utils import is_truthy, secs_to_timestr


class AlertKeywords(LibraryComponent):
    ACCEPT = 'ACCEPT'
    DISMISS = 'DISMISS'
    LEAVE = 'LEAVE'
    _next_alert_action = ACCEPT

    
    @keyword
    def input_text_into_alert(self, text, action=ACCEPT, timeout=None):
        """Types the given ``text`` into an input field in an alert.

        The alert is accepted by default, but that behavior can be controlled
        by using the ``action`` argument same way as with `Handle Alert`.

        ``timeout`` specifies how long to wait for the alert to appear.
        If it is not given, the global default `timeout` is used instead.

        
        """
        alert = self._wait_alert(timeout)
        alert.send_keys(text)
        self._handle_alert(alert, action)

    @keyword
    def alert_should_be_present(self, text='', action=ACCEPT, timeout=None):
        """Verifies that an alert is present and, by default, accepts it.

        Fails if no alert is present. If ``text`` is a non-empty string,
        then it is used to verify alert's message. The alert is accepted
        by default, but that behavior can be controlled by using the
        ``action`` argument same way as with `Handle Alert`.

        ``timeout`` specifies how long to wait for the alert to appear.
        If it is not given, the global default `timeout` is used instead.

       
        """
        message = self.handle_alert(action, timeout)
        if text and text != message:
            raise AssertionError("Alert message should have been '%s' but it "
                                 "was '%s'." % (text, message))

    @keyword
    def alert_should_not_be_present(self, action=ACCEPT, timeout=0):
        """Verifies that no alert is present.

        If the alert actually exists, the ``action`` argument determines
        how it should be handled. By default the alert is accepted, but
        it can be also dismissed or left open the same way as with the
        `Handle Alert` keyword.

        ``timeout`` specifies how long to wait for the alert to appear.
        By default the alert is not waited at all, but a custom time can
        be given if alert may be delayed. See the `time format` section
        for information about the syntax.

        
        """
        try:
            alert = self._wait_alert(timeout)
        except AssertionError:
            return
        text = self._handle_alert(alert, action)
        raise AssertionError("Alert with message '%s' present." % text)

    

    

    

    

    @keyword
    def handle_alert(self, action=ACCEPT, timeout=None):
        """Handles the current alert and returns its message.

        By default the alert is accepted, but this can be controlled
        with the ``action`` argument that supports the following
        case-insensitive values:

        - ``ACCEPT``: Accept the alert i.e. press ``Ok``. Default.
        - ``DISMISS``: Dismiss the alert i.e. press ``Cancel``.
        - ``LEAVE``: Leave the alert open.

        The ``timeout`` argument specifies how long to wait for the alert
        to appear. If it is not given, the global default `timeout` is used
        instead.

        Examples:
        | Handle Alert |                |       | # Accept alert.  |
        | Handle Alert | action=DISMISS |       | # Dismiss alert. |
        | Handle Alert | timeout=10 s   |       | # Use custom timeout and accept alert.  |
        | Handle Alert | DISMISS        | 1 min | # Use custom timeout and dismiss alert. |
        | ${message} = | Handle Alert   |       | # Accept alert and get its message.     |
        | ${message} = | Handle Alert   | LEAVE | # Leave alert open and get its message. |

        
        """
        alert = self._wait_alert(timeout)
        return self._handle_alert(alert, action)

    def _handle_alert(self, alert, action):
        action = action.upper()
        text = ' '.join(alert.text.splitlines())
        if action == self.ACCEPT:
            alert.accept()
        elif action == self.DISMISS:
            alert.dismiss()
        elif action != self.LEAVE:
            raise ValueError("Invalid alert action '%s'." % action)
        return text

    def _wait_alert(self, timeout=None):
        timeout = self.get_timeout(timeout)
        wait = WebDriverWait(self.driver, timeout)
        try:
            return wait.until(EC.alert_is_present())
        except WebDriverException:
            raise AssertionError('Alert not found in %s.'
                                 % secs_to_timestr(timeout))



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# For Python 2
from __future__ import unicode_literals
from __future__ import division
from io import open
from .compat import IS_PYTHON_2, STRING_TYPES

from json import dumps, load, loads
from os import path
from yaml import load as load_yaml

from pygments import highlight, lexers, formatters
from requests.packages.urllib3 import disable_warnings

if IS_PYTHON_2:
    from urlparse import parse_qs, urljoin, urlparse
else:
    from urllib.parse import parse_qs, urljoin, urlparse

from robot.api import logger

from .keywords import Keywords
from .version import __version__


class rbfREST(Keywords):
    """rbfREST is a REST API testing library. 

    = The state =

    The library represents its own state as JSON itself, as an array of objects.
    Together these objects are commonly called instances.

    A single instance always has these three properties:

    - Request data as a JSON object
    - Response data as a JSON object
    - JSON Schema for the above two properties

    For each request-response, as soon as the response has been gotten
    (and the request did not timeout), a new instance is created with these
    properties.

    Request and response schemas are inferred if they are not already
    expected by using expectation keywords. All the validations the library
    implements are based on JSON Schema [http://json-schema.org/draft-07/json-schema-validation.html|draft-07] by default
   

    = The scope =

    All the assertion keywords, `Output` and `Output Schema` are effective
    in the scope of the last instance.

    The scope of the library itself is test suite, meaning the instances
    are persisted in memory until the execution of the test suite is finished,
    regardless whether successfully or not.

    The last request and response is easiest seen with `Output`.
    The output is written to terminal by default, as this is usually faster
    for debugging purposes than finding the right keyword in ``log.html``.

    All instances can be output to a file with `RESTinstances` which can
    be useful for additional logging.
    
    
    = Examples =
    
    HTTP return code validation
    | ${requestHeaders} |  `Load Headers`  | ${EXECDIR}/resources/headers/RequestHeaders.json |
    | `GET` | https://jsonplaceholder.typicode.com/users   | headers=${requestHeaders} |
    | `Assert If Field is Number` | response status   | 200 |
    
    Extract response fields
    | ${requestHeaders} |  `Load Headers`  | ${EXECDIR}/resources/headers/RequestHeaders.json |
    | ${response} | `GET` | https://jsonplaceholder.typicode.com/users?_limit=1   | headers=${requestHeaders} |
    | ${responseField} | `Get Value From Response` | ${response} | $..username |   
    
    Assert response field with value
    | ${requestHeaders} |  `Load Headers`  | ${EXECDIR}/resources/headers/RequestHeaders.json |
    | `GET` | https://jsonplaceholder.typicode.com/users?_limit=1 | headers=${requestHeaders} |
    | `Assert if field is string` | $..username | Bret | 
    
    Assert on multiple API requests
    | ${requestHeaders} |  `Load Headers`  | ${EXECDIR}/resources/headers/RequestHeaders.json |
    | `GET` | https://jsonplaceholder.typicode.com/users?_limit=1   | headers=${requestHeaders} |
    | ${response1} | `Extract Response For Assertions` |
    | `GET` | https://jsonplaceholder.typicode.com/users?_limit=2   | headers=${requestHeaders} |
    | ${response2} | `Extract Response For Assertions` |
    | `Assert if field is string` | $..username | Bret | response=${response1} |
    | `Assert if field is string` | $..username | Antonette | response=${response2} |
    
    """

    ROBOT_LIBRARY_SCOPE = 'TEST SUITE'
    ROBOT_LIBRARY_VERSION = __version__

    # Altogether 24 keywords        context:
    # -------------------------------------------------------
    # 2 setting keywords            next instances
    # 3 expectation keywords        next instances
    # 7 operation keywords          next instance
    # 8 assertion keywords          last instance's schema
    # 4 I/O keywords                the last instance or none
    # -------------------------------------------------------

    def __init__(self, url=None,
                 ssl_verify=True,
                 accept="application/json, */*",
                 content_type="application/json",
                 user_agent="dnb-rf-restapi-library/%s" % (__version__),
                 proxies={},
                 schema={},
                 spec={},
                 instances=[]):
        self.request = {
            'method': None,
            'url': None,
            'scheme': "",
            'netloc': "",
            'path': "",
            'query': {},
            'body': None,
            'headers': {
                'Accept': rbfREST._input_string(accept),
                'Content-Type': rbfREST._input_string(content_type),
                'User-Agent': rbfREST._input_string(user_agent)
            },
            'proxies': rbfREST._input_object(proxies),
            'timeout': [None, None],
            'cert': None,
            'sslVerify': rbfREST._input_ssl_verify(ssl_verify),
            'allowRedirects': True
        }
        if url:
            url = rbfREST._input_string(url)
            if url.endswith('/'):
                url = url[:-1]
            if not url.startswith(("http://", "https://")):
                url = "http://" + url
            url_parts = urlparse(url)
            self.request['scheme'] = url_parts.scheme
            self.request['netloc'] = url_parts.netloc
            self.request['path'] = url_parts.path
        if not self.request['sslVerify']:
            disable_warnings()
        self.schema = {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "title": url,
            "description": None,
            "default": True,
            "examples": [],
            "type": "object",
            "properties": {
                "request": {
                    "type": "object",
                    "properties": {}
                },
                "response": {
                    "type": "object",
                    "properties": {}
                }
            }
        }
        self.schema.update(self._input_object(schema))
        self.spec = {}
        self.spec.update(self._input_object(spec))
        self.instances = self._input_array(instances)


    @staticmethod
    def log_json(json, header="", also_console=True, sort_keys=False):
        json = dumps(json, ensure_ascii=False, indent=4,
                     separators=(',', ': ' ), sort_keys=sort_keys)
        logger.info("%s\n%s" % (header, json))    # no coloring for log.html
        if also_console:
            json_data = highlight(json,
                                  lexers.JsonLexer(),
                                  formatters.TerminalFormatter())
            logger.console("%s\n%s" % (header, json_data), newline=False)
        return json

    @staticmethod
    def _input_boolean(value):
        if isinstance(value, (bool)):
            return value
        try:
            json_value = loads(value)
            if not isinstance(json_value, (bool)):
                raise RuntimeError("Input is not a JSON boolean: %s" % (value))
        except ValueError:
            raise RuntimeError("Input is not valid JSON: %s" % (value))
        return json_value

    @staticmethod
    def _input_integer(value):
        if isinstance(value, (int)):
            return value
        try:
            json_value = loads(value)
            if not isinstance(json_value, (int)):
                raise RuntimeError("Input is not a JSON integer: %s" % (value))
        except ValueError:
            raise RuntimeError("Input is not valid JSON: %s" % (value))
        return json_value

    @staticmethod
    def _input_number(value):
        if isinstance(value, (float, int)):
            return value
        try:
            json_value = loads(value)
            if not isinstance(json_value, (float, int)):
                raise RuntimeError("Input is not a JSON number: %s" % (value))
        except ValueError:
            raise RuntimeError("Input is not valid JSON: %s" % (value))
        return json_value

    @staticmethod
    def _input_string(value):
        if value == "":
            return ""
        if isinstance(value, STRING_TYPES):
            if not value.startswith('"'):
                value = '"' + value
            if not value.endswith('"'):
                value = value + '"'
        try:
            json_value = loads(value)
            if not isinstance(json_value, STRING_TYPES):
                raise RuntimeError("Input is not a JSON string: %s" % (value))
        except ValueError:
            raise RuntimeError("Input not is valid JSON: %s" % (value))
        return json_value


    @staticmethod
    def _input_object(value):
        if isinstance(value, (dict)):
            return value
        try:
            if path.isfile(value):
                json_value = rbfREST._input_json_from_file(value)
            else:
                json_value = loads(value)
            if not isinstance(json_value, (dict)):
                raise RuntimeError("Input or file has no JSON object: %s" % (
                    value))
        except ValueError:
            raise RuntimeError("Input is not valid JSON or a file: %s" % (
                value))
        return json_value

    @staticmethod
    def _input_array(value):
        if isinstance(value, (list)):
            return value
        try:
            if path.isfile(value):
                json_value = rbfREST._input_json_from_file(value)
            else:
                json_value = loads(value)
            if not isinstance(json_value, (list)):
                raise RuntimeError("Input or file has no JSON array: %s" % (
                    value))
        except ValueError:
            raise RuntimeError("Input is not valid JSON or a file: %s" % (
                value))
        return json_value

    @staticmethod
    def _input_json_from_file(path):
        try:
            with open(path, encoding="utf-8") as file:
                return load(file)
        except IOError as e:
            raise RuntimeError("File '%s' cannot be opened:\n%s" % (path, e))
        except ValueError as e:
            try:
                with open(path, encoding="utf-8") as file:
                    return load_yaml(file)
            except ValueError:
                raise RuntimeError("File '%s' is not valid JSON or YAML:\n%s" %
                    (path, e))

    @staticmethod
    def _input_json_as_string(string):
        return loads(string)

    @staticmethod
    def _input_json_from_non_string(value):
        try:
            return rbfREST._input_json_as_string(dumps(value, ensure_ascii=False))
        except ValueError:
            raise RuntimeError("Input is not valid JSON: %s" % (value))

    @staticmethod
    def _input_client_cert(value):
        if value is None or value == "null":
            return None
        if isinstance(value, STRING_TYPES):
            return value
        if isinstance(value, (list)):
            if len(value) != 2:
                raise RuntimeError("Client cert given as a (Python) list, " +
                    "must have length of 2: %s" % (value))
            return value
        try:
            value = loads(value)
            if not isinstance(value, STRING_TYPES + (list)):
                raise RuntimeError("Input is not a JSON string " +
                    "or a list: %s" + (value))
        except ValueError:
            raise RuntimeError("Input is not a JSON string " +
                "or an array: %s " % (value))
        if isinstance(value, (list)):
            if len(value) != 2:
                raise RuntimeError("Client cert given as a JSON array, " +
                    "must have length of 2: %s" % (value))
        return value

    @staticmethod
    def _input_ssl_verify(value):
        try:
            return rbfREST._input_boolean(value)
        except RuntimeError:
            value = rbfREST._input_string(value)
            if not path.isfile(value):
                raise RuntimeError("SSL verify option is not " +
                    "a Python or a JSON boolean or a path to an existing " +
                    "CA bundle file: %s" % (value))
            return value

    @staticmethod
    def _input_timeout(value):
        if isinstance(value, (int, float)):
            return [value, value]
        if isinstance(value, (list)):
            if len(value) != 2:
                raise RuntimeError("Timeout given as a (Python) list, " +
                    "must have length of 2: %s" % (value))
            return value
        try:
            value = loads(value)
            if not isinstance(value, (int, float, list)):
                raise RuntimeError("Input is not a JSON integer, " +
                    "number or a list: %s" % (value))
        except ValueError:
            raise RuntimeError("Input is not valid JSON: %s" % (value))
        if isinstance(value, (list)):
            if len(value) != 2:
                raise RuntimeError("Timeout given as a JSON array, " +
                    "must have length of 2: %s" % (value))
            else:
                return value
        return [value, value]
